"""Service mode for MPF."""
import subprocess
import os
from collections import namedtuple

from typing import List

from mpf.core.async_mode import AsyncMode
from mpf.core.switch_controller import MonitoredSwitchChange
from mpf.core.utility_functions import Util

ServiceMenuEntry = namedtuple("ServiceMenuEntry", ["label", "callback"])


class Service(AsyncMode):

    """The service mode."""

    __slots__ = ["_update_script"]

    def __init__(self, *args, **kwargs):
        """Initialize service mode."""
        super().__init__(*args, **kwargs)
        self._update_script = None

    @staticmethod
    def get_config_spec():
        """Add validation for mode."""
        return '''
door_opened_events: list|str|service_door_opened
door_closed_events: list|str|service_door_closed
high_power_on_events: list|str|service_power_on
high_power_off_events: list|str|service_power_off
enter_events: list|str|sw_service_enter_active
esc_events: list|str|sw_service_esc_active
up_events: list|str|sw_service_up_active
down_events: list|str|sw_service_down_active
software_update: single|bool|False
software_update_script: single|str|None
'''

    async def _service_mode_exit(self):
        await self.machine.service.stop_service()

    def _get_key(self):
        return Util.race({
            self.machine.events.wait_for_any_event(self.config['mode_settings']['esc_events']): "ESC",
            self.machine.events.wait_for_any_event(self.config['mode_settings']['enter_events']): "ENTER",
            self.machine.events.wait_for_any_event(self.config['mode_settings']['up_events']): "UP",
            self.machine.events.wait_for_any_event(self.config['mode_settings']['down_events']): "DOWN",
        })

    async def _run(self):
        while True:
            # wait for key
            key = await self._get_key()

            if key == "ENTER":
                # start main menu
                await self._start_main_menu()
            elif key == "UP":
                # post event for mc to increase volume
                self.machine.events.post("master_volume_increase")
                '''event: master_volume_increase

                desc: Increase the master volume of the audio system.
                '''
            elif key == "DOWN":
                # post event for mc to decrease volume
                self.machine.events.post("master_volume_decrease")
                '''event: master_volume_decrease

                desc: Decrease the master volume of the audio system.
                '''

    async def _start_main_menu(self):
        self.machine.service.start_service()
        self.machine.events.post("service_main_menu")
        await self._service_mode_main_menu()

        await self._service_mode_exit()

    def _update_main_menu(self, items: List[ServiceMenuEntry], position: int):
        self.machine.events.post("service_menu_deselected")
        self.machine.events.post("service_menu_show")
        self.machine.events.post("service_menu_selected", label=items[position].label)

    def _load_menu_entries(self) -> List[ServiceMenuEntry]:
        """Return the menu items with label and callback."""
        # If you want to add menu entries overload the mode and this method.
        entries = [
            ServiceMenuEntry("Diagnostics Menu", self._diagnostics_menu),
            ServiceMenuEntry("Audits Menu", self._audits_menu),
            ServiceMenuEntry("Adjustments Menu", self._adjustments_menu),
            ServiceMenuEntry("Utilities Menu", self._utilities_menu),

        ]
        return entries

    async def _service_mode_main_menu(self):
        """Show main menu."""
        await self._make_menu(self._load_menu_entries())
        # hide slides on exit
        self.machine.events.post("service_menu_hide")

    # Diagnostics
    def _load_diagnostic_menu_entries(self) -> List[ServiceMenuEntry]:
        """Return the diagnostics menu items with label and callback."""
        return [
            ServiceMenuEntry("Switch Menu", self._diagnostics_switch_menu),
            ServiceMenuEntry("Coil Menu", self._diagnostics_coil_menu),
            ServiceMenuEntry("Light Menu", self._diagnostics_light_menu),
        ]

    async def _diagnostics_menu(self):
        await self._make_menu(self._load_diagnostic_menu_entries())

    def _load_diagnostic_switch_menu_entries(self) -> List[ServiceMenuEntry]:
        """Return the switch menu items with label and callback."""
        return [
            ServiceMenuEntry("Switch Edge Test", self._switch_test_menu),
        ]

    async def _diagnostics_switch_menu(self):
        await self._make_menu(self._load_diagnostic_switch_menu_entries())

    def _load_diagnostic_coil_menu_entries(self) -> List[ServiceMenuEntry]:
        """Return the coil menu items with label and callback."""
        return [
            ServiceMenuEntry("Single Coil Test", self._coil_test_menu),
        ]

    async def _diagnostics_coil_menu(self):
        await self._make_menu(self._load_diagnostic_coil_menu_entries())

    def _load_diagnostic_light_menu_entries(self) -> List[ServiceMenuEntry]:
        """Return the light menu items with label and callback."""
        return [
            ServiceMenuEntry("Single Light Test", self._light_test_menu),
        ]

    async def _diagnostics_light_menu(self):
        await self._make_menu(self._load_diagnostic_light_menu_entries())

    # Audits
    def _load_audit_menu_entries(self) -> List[ServiceMenuEntry]:
        """Return the audit menu items with label and callback."""
        return [
            # ServiceMenuEntry("Earning Audits", None),
            ServiceMenuEntry("Standard Audits", self._audit_standard_menu),
            ServiceMenuEntry("Feature Audits", self._audit_feature_menu),
        ]

    async def _audits_menu(self):
        await self._make_menu(self._load_audit_menu_entries())

    # Adjustments
    def _load_adjustments_menu_entries(self) -> List[ServiceMenuEntry]:
        """Return the adjustments menu items with label and callback."""
        return [
            ServiceMenuEntry("Standard Adjustments", self._settings_menu),
        ]

    async def _adjustments_menu(self):
        await self._make_menu(self._load_adjustments_menu_entries())

    # Utilities
    def _load_utilities_menu_entries(self) -> List[ServiceMenuEntry]:
        """Return the utilities menu items with label and callback."""
        entries = [
            ServiceMenuEntry("Reset Menu", self._utilities_reset_menu),
        ]

        if self.config['mode_settings']['software_update']:
            update_file_path = self.config['mode_settings']['software_update_script']
            if not update_file_path:
                raise AssertionError("Please configure software_update_script to enable software_update in "
                                     "service mode.")

            if not os.path.isabs(update_file_path):
                update_file_path = os.path.join(self.machine.machine_path, update_file_path)

            if os.path.isfile(update_file_path):
                self._update_script = update_file_path
                entries.append(ServiceMenuEntry("Software Update", self._software_update))

        return entries

    async def _utilities_menu(self):
        await self._make_menu(self._load_utilities_menu_entries())

    def _load_utilities_reset_menu_entries(self) -> List[ServiceMenuEntry]:
        """Return the utilities reset menu items with label and callback."""
        return [
            ServiceMenuEntry("Reset Coin Audits", self._utilities_reset_coin_audits),
            ServiceMenuEntry("Reset Game Audits", self._utilities_reset_game_audits),
            ServiceMenuEntry("Reset High Scores", self._utilities_reset_high_scores),
            ServiceMenuEntry("Reset Credits", self._utilities_reset_credits),
            ServiceMenuEntry("Reset to Factory Settings", self._utilities_reset_to_factory_settings),
        ]

    async def _make_option_slide(self, title, question, options, warning):
        """Show service_options_slide, provide options and return the selected option."""
        position = 0

        while True:
            self.machine.events.post("service_options_slide_start", title=title, question=question,
                                     option=options[position], warning=warning)
            key = await self._get_key()
            if key == 'ESC':
                return None
            if key == 'UP':
                position += 1
                if position >= len(options):
                    position = 0
            elif key == 'DOWN':
                position -= 1
                if position < 0:
                    position = len(options) - 1
            elif key == 'ENTER':
                # select and return option
                self.machine.events.post("service_options_slide_stop")
                return options[position]

    async def _utilities_reset_menu(self):
        await self._make_menu(self._load_utilities_reset_menu_entries())

    async def _utilities_reset_coin_audits(self):
        selection = await self._make_option_slide("Reset earning audits", "Perform coin reset?", ["no", "yes"],
                                                  "THIS CANNOT BE UNDONE")
        if selection == "yes":
            self.machine.events.post("earnings_reset")

    async def _utilities_reset_game_audits(self):
        selection = await self._make_option_slide("Auditor Reset", "Reset Game Audits?", ["no", "yes"],
                                                  "THIS CANNOT BE UNDONE")
        if selection == "yes":
            self.machine.events.post("auditor_reset")

    async def _utilities_reset_high_scores(self):
        selection = await self._make_option_slide("High Score Reset", "Remove Highscores?", ["no", "yes"],
                                                  "THIS CANNOT BE UNDONE")
        if selection == "yes":
            self.machine.events.post("high_scores_reset")

    async def _utilities_reset_credits(self):
        selection = await self._make_option_slide("Reset credits", "Remove all credits?", ["no", "yes"],
                                                  "THIS CANNOT BE UNDONE")
        if selection == "yes":
            self.machine.events.post("credits_reset")

    async def _utilities_reset_to_factory_settings(self):
        selection = await self._make_option_slide("Factory Reset", "Reset to factory setting?", ["no", "yes"],
                                                  "THIS CANNOT BE UNDONE")
        if selection == "yes":
            self.machine.events.post("factory_reset")

    async def _software_update(self):
        run_update = False
        self._update_software_update_slide(run_update)

        while True:
            key = await self._get_key()
            if key == 'ESC':
                break
            if key in ('UP', 'DOWN'):
                run_update = not run_update
                self._update_software_update_slide(run_update)
            elif key == 'ENTER':
                # perform update
                if run_update:
                    self.machine.events.post("service_software_update_start")
                    subprocess.Popen([self._update_script])
                    self.machine.stop("Software Update")

        self.machine.events.post("service_software_update_stop")

    def _update_software_update_slide(self, run_update):
        self.machine.events.post("service_software_update_choice", run_update="Yes" if run_update else "No")

    async def _make_menu(self, items):
        """Show a menu by controlling slides via events and executing callbacks."""
        if not items:
            # do not crash on empty menu
            return
        position = 0
        self._update_main_menu(items, position)

        while True:
            key = await self._get_key()
            if key == 'ESC':
                return
            if key == 'UP':
                position += 1
                if position >= len(items):
                    position = 0
                self._update_main_menu(items, position)
            elif key == 'DOWN':
                position -= 1
                if position < 0:
                    position = len(items) - 1
                self._update_main_menu(items, position)
            elif key == 'ENTER':
                # call submenu
                await items[position].callback()
                self._update_main_menu(items, position)

    def _switch_monitor(self, change: MonitoredSwitchChange):
        if change.state:
            state_string = "active"
            self.machine.events.post("service_switch_hit")
        else:
            state_string = "inactive"

        self.machine.events.post("service_switch_test_start",
                                 switch_name=change.name,
                                 switch_num=change.num,
                                 switch_label=change.label,
                                 switch_state=state_string)

    async def _switch_test_menu(self):
        self.machine.switch_controller.add_monitor(self._switch_monitor)
        self.machine.events.post("service_switch_test_start",
                                 switch_name="", switch_state="", switch_num="", switch_label="")
        await self.machine.events.wait_for_any_event(self.config['mode_settings']['esc_events'])
        self.machine.events.post("service_switch_test_stop")
        self.machine.switch_controller.remove_monitor(self._switch_monitor)

    def _update_coil_slide(self, items, position):
        board, coil = items[position]
        self.machine.events.post("service_coil_test_start",
                                 board_name=board,
                                 coil_name=coil.name,
                                 coil_label=coil.config['label'],
                                 coil_num=coil.hw_driver.number)

    async def _coil_test_menu(self):
        position = 0
        items = self.machine.service.get_coil_map()

        # do not crash if no coils are configured
        if not items:   # pragma: no cover
            return

        self._update_coil_slide(items, position)

        while True:
            key = await self._get_key()
            if key == 'ESC':
                break
            if key == 'UP':
                position += 1
                if position >= len(items):
                    position = 0
                self._update_coil_slide(items, position)
            elif key == 'DOWN':
                position -= 1
                if position < 0:
                    position = len(items) - 1
                self._update_coil_slide(items, position)
            elif key == 'ENTER':
                # pulse coil
                items[position].coil.pulse()

        self.machine.events.post("service_coil_test_stop")

    def _update_light_slide(self, items, position, color):
        board, light = items[position]
        self.machine.events.post("service_light_test_start",
                                 board_name=board,
                                 light_name=light.name,
                                 light_label=light.config['label'],
                                 light_num=light.config['number'],
                                 test_color=color)

    async def _light_test_menu(self):
        position = 0
        color_position = 0
        colors = ["white", "red", "green", "blue", "yellow"]
        items = self.machine.service.get_light_map()

        # do not crash if no lights are configured
        if not items:   # pragma: no cover
            return

        self._update_light_slide(items, position, colors[color_position])

        while True:
            self._update_light_slide(items, position, colors[color_position])
            items[position].light.color(colors[color_position], key="service", priority=1000000)

            key = await self._get_key()
            items[position].light.remove_from_stack_by_key("service")
            if key == 'ESC':
                break
            if key == 'UP':
                position += 1
                if position >= len(items):
                    position = 0
            elif key == 'DOWN':
                position -= 1
                if position < 0:
                    position = len(items) - 1
            elif key == 'ENTER':
                # change color
                color_position += 1
                if color_position >= len(colors):
                    color_position = 0

        self.machine.events.post("service_light_test_stop")

    def _update_settings_slide(self, items, position, is_change=False):
        setting = items[position]
        label = self.machine.settings.get_setting_value_label(setting.name)
        event = "service_settings_{}".format("edit" if is_change else "start")
        self.machine.events.post(event,
                                 settings_label=setting.label,
                                 value_label=label)

    async def _settings_menu(self):
        position = 0
        items = self.machine.settings.get_settings()

        # do not crash if no settings
        if not items:   # pragma: no cover
            return

        self._update_settings_slide(items, position)

        while True:
            key = await self._get_key()
            if key == 'ESC':
                break
            if key == 'UP':
                position += 1
                if position >= len(items):
                    position = 0
                self._update_settings_slide(items, position)
            elif key == 'DOWN':
                position -= 1
                if position < 0:
                    position = len(items) - 1
                self._update_settings_slide(items, position)
            elif key == 'ENTER':
                # change setting
                await self._settings_change(items, position)

        self.machine.events.post("service_settings_stop")

    async def _settings_change(self, items, position):
        self._update_settings_slide(items, position)

        values = list(items[position].values.keys())
        value_position = values.index(self.machine.settings.get_setting_value(items[position].name))
        self._update_settings_slide(items, position, is_change=True)

        while True:
            key = await self._get_key()
            if key == 'ESC':
                self._update_settings_slide(items, position)
                break
            if key == 'UP':
                value_position += 1
                if value_position >= len(values):
                    value_position = 0
                self.machine.settings.set_setting_value(items[position].name, values[value_position])
                self._update_settings_slide(items, position, is_change=True)
            elif key == 'DOWN':
                value_position -= 1
                if value_position < 0:
                    value_position = len(values) - 1
                self.machine.settings.set_setting_value(items[position].name, values[value_position])
                self._update_settings_slide(items, position, is_change=True)

    # ENTER hit when on Standard Audit sub menu, post events for user to show slides
    async def _audit_standard_menu(self):
        self.machine.events.post("service_audit_standard_start")

        while True:
            key = await self._get_key()
            if key == 'ESC':
                break
            if key == 'ENTER':
                pass

        self.machine.events.post("service_audit_standard_stop")

    # ENTER hit when on Features Audit sub menu, post events for user to show slides
    async def _audit_feature_menu(self):
        self.machine.events.post("service_audit_feature_start")

        while True:
            key = await self._get_key()
            if key == 'ESC':
                break
            if key == 'ENTER':
                pass

        self.machine.events.post("service_audit_feature_stop")



from mpf.tests.MpfTestCase import MpfTestCase, MagicMock


class TestCarouselMode(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/carousel/'

    def _start_game(self):
        self.machine.playfield.add_ball = MagicMock()
        self.machine.ball_controller.num_balls_known = 3
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertIsNotNone(self.machine.game)

    def _stop_game(self):
        # stop game
        self.assertIsNotNone(self.machine.game)
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertIsNone(self.machine.game)

    def testBlockingCarousel(self):
        self.mock_event("blocking_carousel_item1_highlighted")
        self.mock_event("blocking_carousel_item2_highlighted")
        self.mock_event("blocking_carousel_item3_highlighted")
        self.mock_event("flipper_cancel")
        
        self._start_game()
        self.post_event("start_mode4")

        self.assertIn(self.machine.modes["blocking_carousel"], self.machine.mode_controller.active_modes)
        self.assertEqual(1, self._events["blocking_carousel_item1_highlighted"])
        self.assertEqual(0, self._events["blocking_carousel_item2_highlighted"]) 
        self.post_event("s_flipper_right_active")
        self.post_event("s_flipper_right_inactive")
        self.assertEqual(1, self._events["blocking_carousel_item2_highlighted"])
        self.assertEqual(0, self._events["blocking_carousel_item3_highlighted"]) 
        self.assertEqual(0, self._events["flipper_cancel"]) 
        self.post_event("s_flipper_right_active")
        self.post_event("s_flipper_left_active")
        self.post_event("flipper_cancel")
        self.post_event("s_flipper_right_inactive")
        self.post_event("s_flipper_left_inactive")
        self.assertEqual(1, self._events["flipper_cancel"])
        self.assertEqual(1, self._events["blocking_carousel_item1_highlighted"])
        self.assertEqual(1, self._events["blocking_carousel_item2_highlighted"]) 
        self.assertEqual(0, self._events["blocking_carousel_item3_highlighted"]) 
        self.post_event("both_flippers_inactive")
        self.post_event("s_flipper_right_inactive")
        self.assertEqual(1, self._events["blocking_carousel_item3_highlighted"]) 

        # Restart the mode to ensure that the block is cleared
        self.post_event("flipper_cancel")
        self.post_event("stop_mode4")
        self.advance_time_and_run()
        self.post_event("start_mode4")
        self.post_event("s_flipper_right_inactive")
        self.assertEqual(2, self._events["blocking_carousel_item2_highlighted"], 
            "item2_highlighted should be called when a blocked mode restarts") 

    def testConditionalCarousel(self):
        self.mock_event("conditional_carousel_item1_highlighted")
        self.mock_event("conditional_carousel_item2_highlighted")
        self.mock_event("conditional_carousel_item3_highlighted")
        self.mock_event("conditional_carousel_item4_highlighted")

        self._start_game()

        # Start the mode without any conditions true
        self.post_event("start_mode3")
        self.assertIn(self.machine.modes["conditional_carousel"], self.machine.mode_controller.active_modes)
        self.assertEqual(1, self._events["conditional_carousel_item1_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item2_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item3_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item4_highlighted"])
        self.post_event("next")
        self.assertEqual(2, self._events["conditional_carousel_item1_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item2_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item3_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item4_highlighted"])
        self.post_event("next")
        self.assertEqual(3, self._events["conditional_carousel_item1_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item2_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item3_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item4_highlighted"])
        self.post_event("stop_mode3")

        # Reset the count for item 1
        self.mock_event("conditional_carousel_item1_highlighted")
        # Start the mode with a player variable condition
        self.machine.game.player["show_item4"] = True
        self.post_event("start_mode3")
        self.assertEqual(1, self._events["conditional_carousel_item1_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item2_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item3_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item4_highlighted"])
        self.post_event("next")
        self.assertEqual(1, self._events["conditional_carousel_item1_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item2_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item3_highlighted"])
        self.assertEqual(1, self._events["conditional_carousel_item4_highlighted"])
        self.post_event("next")
        self.assertEqual(2, self._events["conditional_carousel_item1_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item2_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item3_highlighted"])
        self.assertEqual(1, self._events["conditional_carousel_item4_highlighted"])
        self.post_event("stop_mode3")

        # Reset the count for items 1 and 4
        self.mock_event("conditional_carousel_item1_highlighted")
        self.mock_event("conditional_carousel_item4_highlighted")
        # Start the mode with a machine variable condition
        self.machine.variables.set_machine_var("player2_score", 500000)
        self.machine.game.player["show_item4"] = False
        self.post_event("start_mode3")
        self.assertEqual(1, self._events["conditional_carousel_item1_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item2_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item3_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item4_highlighted"])
        self.post_event("next")
        self.assertEqual(1, self._events["conditional_carousel_item1_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item2_highlighted"])
        self.assertEqual(1, self._events["conditional_carousel_item3_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item4_highlighted"])
        self.post_event("next")
        self.assertEqual(2, self._events["conditional_carousel_item1_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item2_highlighted"])
        self.assertEqual(1, self._events["conditional_carousel_item3_highlighted"])
        self.assertEqual(0, self._events["conditional_carousel_item4_highlighted"])
        self.post_event("stop_mode3")

        # The mode shouldn't start if all conditions are false (i.e. no items)
        self.mock_event("conditional_carousel_items_empty")
        self.machine.game.player["hide_item1"] = "truthy"
        self.machine.variables.set_machine_var("player2_score", 0)
        self.post_event("start_mode3")
        self.assertEqual(1, self._events["conditional_carousel_items_empty"])
        self.assertNotIn(self.machine.modes["conditional_carousel"], self.machine.mode_controller.active_modes)

    def testExtraBall(self):
        self.mock_event("carousel_item1_highlighted")
        self.mock_event("carousel_item1_selected")
        self.mock_event("carousel_item2_highlighted")
        self.mock_event("carousel_item2_selected")
        self.mock_event("carousel_item3_highlighted")
        self.mock_event("carousel_item3_selected")

        # start game
        self._start_game()
        # start mode
        self.post_event("start_mode1")
        self.assertIn(self.machine.modes["carousel"], self.machine.mode_controller.active_modes)

        self.assertEqual(1, self._events["carousel_item1_highlighted"])
        self.assertEqual(0, self._events["carousel_item2_highlighted"])
        self.assertEqual(0, self._events["carousel_item3_highlighted"])

        self.post_event("next")
        self.assertEqual(1, self._events["carousel_item1_highlighted"])
        self.assertEqual(1, self._events["carousel_item2_highlighted"])
        self.assertEqual(0, self._events["carousel_item3_highlighted"])

        self.post_event("next")
        self.assertEqual(1, self._events["carousel_item1_highlighted"])
        self.assertEqual(1, self._events["carousel_item2_highlighted"])
        self.assertEqual(1, self._events["carousel_item3_highlighted"])

        self.post_event("next")
        self.assertEqual(2, self._events["carousel_item1_highlighted"])
        self.assertEqual(1, self._events["carousel_item2_highlighted"])
        self.assertEqual(1, self._events["carousel_item3_highlighted"])

        self.post_event("previous2")
        self.assertEqual(2, self._events["carousel_item1_highlighted"])
        self.assertEqual(1, self._events["carousel_item2_highlighted"])
        self.assertEqual(2, self._events["carousel_item3_highlighted"])

        self.post_event("previous")
        self.assertEqual(2, self._events["carousel_item1_highlighted"])
        self.assertEqual(2, self._events["carousel_item2_highlighted"])
        self.assertEqual(2, self._events["carousel_item3_highlighted"])

        self.post_event("select")
        self.assertEqual(0, self._events["carousel_item1_selected"])
        self.assertEqual(1, self._events["carousel_item2_selected"])
        self.assertEqual(0, self._events["carousel_item3_selected"])

        self.assertNotIn(self.machine.modes["carousel"], self.machine.mode_controller.active_modes)



"""A generic system11 driver overlay.

This is based on the Snux platform to generically support all kinds of System11 platforms.
"""
from typing import Any, Optional, Set, Tuple, Dict

from mpf.core.machine import MachineController
from mpf.core.platform import DriverPlatform, DriverConfig, SwitchSettings, DriverSettings, RepulseSettings, \
    SwitchPlatform, SwitchConfig

from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings

from mpf.core.delays import DelayManager

MYPY = False
if MYPY:   # pragma: no cover
    class SwitchDriverPlatform(DriverPlatform, SwitchPlatform):     # noqa
        pass


# pylint: disable-msg=too-many-instance-attributes
class System11OverlayPlatform(DriverPlatform, SwitchPlatform):

    """Overlay platform to drive system11 machines using a WPC controller."""

    __slots__ = ["delay", "platform", "system11_config", "a_side_queue", "c_side_queue",
                 "a_side_done_time", "c_side_done_time", "drivers_holding_a_side", "drivers_holding_c_side",
                 "a_side_enabled", "c_side_enabled", "ac_relay_in_transition", "prefer_a_side", "drivers"]

    def __init__(self, machine: MachineController) -> None:
        """Initialise the board."""
        super().__init__(machine)

        self.delay = DelayManager(machine)

        self.platform = None            # type: Optional[SwitchDriverPlatform]

        self.system11_config = None     # type: Any

        self.a_side_queue = \
            set()   # type: Set[Tuple[DriverPlatformInterface, Optional[PulseSettings], Optional[HoldSettings]]]
        self.c_side_queue = \
            set()   # type: Set[Tuple[DriverPlatformInterface, Optional[PulseSettings], Optional[HoldSettings]]]

        self.a_side_done_time = 0
        self.c_side_done_time = 0
        self.drivers_holding_a_side = set()     # type: Set[DriverPlatformInterface]
        self.drivers_holding_c_side = set()     # type: Set[DriverPlatformInterface]
        self.a_side_enabled = True
        self.c_side_enabled = False
        self.drivers = {}               # type: Dict[str, DriverPlatformInterface]

        self.ac_relay_in_transition = False
        # Specify whether the AC relay should favour the A or C side when at rest.
        # Typically during a game the 'C' side should be preferred, since that is
        # normally where the flashers are which need a quick response without having to wait on the relay.
        # At game over though, it should prefer the 'A' side so that the relay isn't permanently energised.
        self.prefer_a_side = True

    def stop(self):
        """Stop the overlay. Nothing to do here because stop is also called on parent platform."""

    @property
    def a_side_busy(self):
        """Return if A side cannot be switches off right away."""
        return self.drivers_holding_a_side or self.a_side_done_time > self.machine.clock.get_time() or self.a_side_queue

    @property
    def c_side_active(self):
        """Return if C side cannot be switches off right away."""
        return self.drivers_holding_c_side or self.c_side_done_time > self.machine.clock.get_time()

    @property
    def c_side_busy(self):
        """Return if C side cannot be switches off right away."""
        return self.drivers_holding_c_side or self.c_side_done_time > self.machine.clock.get_time() or self.c_side_queue

    @property
    def a_side_active(self):
        """Return if A side cannot be switches off right away."""
        return self.drivers_holding_a_side or self.a_side_done_time > self.machine.clock.get_time()

    def _null_log_handler(self, *args, **kwargs):
        pass

    async def initialize(self):
        """Automatically called by the Platform class after all the core modules are loaded."""
        # load coil platform
        self.platform = self.machine.get_platform_sections(
            "platform", getattr(self.machine.config.get('system11', {}), 'platform', None))

        # we have to wait for coils to be initialized
        self.machine.events.add_handler("init_phase_1", self._initialize)

    def _initialize(self, **kwargs):
        del kwargs
        self._validate_config()

        self.configure_logging('Platform.System11', self.system11_config['console_log'],
                               self.system11_config['file_log'])

        self.log.debug("Configuring A/C Select Relay for driver %s",
                       self.system11_config['ac_relay_driver'].name)

        self.system11_config['ac_relay_driver'].get_and_verify_hold_power(1.0)

        self.log.debug("Configuring A/C Select Relay transition delay for "
                       "%sms", self.system11_config['ac_relay_delay_ms'])

        self.machine.events.add_handler(self.system11_config['prefer_a_side_event'], self._prefer_a_side)
        self.log.info("Configuring System11 driver to prefer A side on event %s",
                      self.system11_config['prefer_a_side_event'])

        self.machine.events.add_handler(self.system11_config['prefer_c_side_event'], self._prefer_c_side)
        self.log.info("Configuring System11 driver to prefer C side on event %s",
                      self.system11_config['prefer_c_side_event'])

    def _prefer_a_side(self, **kwargs):
        del kwargs
        self.prefer_a_side = True
        self._enable_a_side()

    def _prefer_c_side(self, **kwargs):
        del kwargs
        self.prefer_a_side = False
        self._enable_c_side()

    def _validate_config(self):
        self.system11_config = self.machine.config_validator.validate_config(
            'system11', self.machine.config.get('system11', {}))

    def tick(self):
        """System11 main loop.

        Called based on the timer_tick event.
        """
        if self.prefer_a_side:
            if self.a_side_queue:
                self._service_a_side()
            elif self.c_side_queue:
                self._service_c_side()
            elif self.c_side_enabled and not self.c_side_active:
                self._enable_a_side()
        else:
            if self.c_side_queue:
                self._service_c_side()
            elif self.a_side_queue:
                self._service_a_side()
            elif self.a_side_enabled and not self.a_side_active:
                self._enable_c_side()

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict):
        """Configure switch on system11 overlay."""
        return self.platform.configure_switch(number, config, platform_config)

    async def get_hw_switch_states(self):
        """Get initial hardware state."""
        return await self.platform.get_hw_switch_states()

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict):
        """Configure a driver on the system11 overlay.

        Args:
        ----
            config: Driver config dict
            number: Number of the driver.
            platform_settings: Platform specific config.
        """
        assert self.platform is not None
        orig_number = number

        if number and (number.lower().endswith('a') or number.lower().endswith('c')):

            side = number[-1:].upper()
            number = number[:-1]

            # only configure driver once
            if number not in self.drivers:
                self.drivers[number] = self.platform.configure_driver(config, number, platform_settings)

            system11_driver = System11Driver(orig_number, self.drivers[number], self, side)

            return system11_driver

        return self.platform.configure_driver(config, number, platform_settings)

    @staticmethod
    def _check_if_driver_is_capable_for_rule(driver: DriverPlatformInterface):
        """Check if driver is capable for rule and bail out with an exception if not."""
        number = driver.number
        if number and (number.lower().endswith('a') or number.lower().endswith('c')):
            raise AssertionError("Received a request to set a hardware rule for a System11 driver {}. "
                                 "This is not supported.".format(driver))

    def set_pulse_on_hit_and_release_rule(self, enable_switch, coil):
        """Configure a rule for a driver on the system11 overlay.

        Will pass the call onto the parent platform if the driver is not on A/C relay.
        """
        self._check_if_driver_is_capable_for_rule(coil.hw_driver)

        self.platform.set_pulse_on_hit_and_release_rule(enable_switch, coil)

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch, coil):
        """Configure a rule for a driver on the system11 overlay.

        Will pass the call onto the parent platform if the driver is not on A/C relay.
        """
        self._check_if_driver_is_capable_for_rule(coil.hw_driver)

        self.platform.set_pulse_on_hit_and_enable_and_release_rule(enable_switch, coil)

    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings, eos_switch: SwitchSettings,
                                                      coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Configure a rule for a driver on the system11 overlay.

        Will pass the call onto the parent platform if the driver is not on A/C relay.
        """
        self._check_if_driver_is_capable_for_rule(coil.hw_driver)

        self.platform.set_pulse_on_hit_and_release_and_disable_rule(enable_switch, eos_switch, coil, repulse_settings)

    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings,
                                                                 coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Configure a rule for a driver on the system11 overlay.

        Will pass the call onto the parent platform if the driver is not on A/C relay.
        """
        self._check_if_driver_is_capable_for_rule(coil.hw_driver)

        self.platform.set_pulse_on_hit_and_enable_and_release_and_disable_rule(enable_switch, eos_switch, coil,
                                                                               repulse_settings)

    def set_pulse_on_hit_rule(self, enable_switch, coil):
        """Configure a rule on the system11 overlay.

        Will pass the call onto the parent platform if the driver is not on A/C relay.
        """
        self._check_if_driver_is_capable_for_rule(coil.hw_driver)

        self.platform.set_pulse_on_hit_rule(enable_switch, coil)

    def clear_hw_rule(self, switch, coil):
        """Clear a rule for a driver on the system11 overlay."""
        self._check_if_driver_is_capable_for_rule(coil.hw_driver)

        self.platform.clear_hw_rule(switch, coil)

    def driver_action(self, driver, pulse_settings: Optional[PulseSettings], hold_settings: Optional[HoldSettings],
                      side: str):
        """Add a driver action for a switched driver to the queue (for either the A-side or C-side queue).

        Args:
        ----
            driver: A reference to the original platform class Driver instance.
            pulse_settings: Settings for the pulse or None
            hold_settings:Settings for hold or None
            side: Whatever the driver is on A or C side.

        This action will be serviced immediately if it can, or ASAP otherwise.
        """
        if self.prefer_a_side:
            if side == "A":
                self.a_side_queue.add((driver, pulse_settings, hold_settings))
                self._service_a_side()
            elif side == "C":
                self.c_side_queue.add((driver, pulse_settings, hold_settings))
                if not self.ac_relay_in_transition and not self.a_side_busy:
                    self._service_c_side()
            else:
                raise AssertionError("Invalid side {}".format(side))
        else:
            if side == "C":
                self.c_side_queue.add((driver, pulse_settings, hold_settings))
                self._service_c_side()
            elif side == "A":
                self.a_side_queue.add((driver, pulse_settings, hold_settings))
                if not self.ac_relay_in_transition and not self.c_side_busy:
                    self._service_a_side()
            else:
                raise AssertionError("Invalid side {}".format(side))

    def _enable_ac_relay(self):
        self.system11_config['ac_relay_driver'].enable()
        self.ac_relay_in_transition = True
        self.a_side_enabled = False
        self.c_side_enabled = False
        self.delay.add(ms=self.system11_config['ac_relay_delay_ms'],
                       callback=self._c_side_enabled,
                       name='enable_ac_relay')

    def _disable_ac_relay(self):
        self.system11_config['ac_relay_driver'].disable()
        self.ac_relay_in_transition = True
        self.a_side_enabled = False
        self.c_side_enabled = False
        self.delay.add(ms=self.system11_config['ac_relay_delay_ms'],
                       callback=self._a_side_enabled,
                       name='disable_ac_relay')

    # -------------------------------- A SIDE ---------------------------------

    def _enable_a_side(self):
        if self.prefer_a_side:
            if not self.a_side_enabled and not self.ac_relay_in_transition:

                if self.c_side_active:
                    self._disable_all_c_side_drivers()
                    self._disable_ac_relay()
                    self.delay.add(ms=self.system11_config['ac_relay_delay_ms'],
                                   callback=self._enable_a_side,
                                   name='enable_a_side')
                    return

                if self.c_side_enabled:
                    self._disable_ac_relay()

                else:
                    self._a_side_enabled()
        else:
            if (not self.ac_relay_in_transition and
                    not self.a_side_enabled and
                    not self.c_side_busy):
                self._disable_ac_relay()

            elif self.a_side_enabled and self.a_side_queue:
                self._service_a_side()

    def _a_side_enabled(self):
        self.ac_relay_in_transition = False
        if self.prefer_a_side:
            self.a_side_enabled = True
            self.c_side_enabled = False
            self._service_a_side()
        else:

            if self.c_side_queue:
                self._enable_c_side()
                return

            self.c_side_enabled = False
            self.a_side_enabled = True
            self._service_a_side()

    def _service_a_side(self):
        if not self.a_side_queue:
            return

        if not self.a_side_enabled:
            self._enable_a_side()
            return

        while self.a_side_queue:
            driver, pulse_settings, hold_settings = self.a_side_queue.pop()

            if hold_settings is None and pulse_settings:
                driver.pulse(pulse_settings)
                self.a_side_done_time = max(self.a_side_done_time,
                                            self.machine.clock.get_time() + (pulse_settings.duration / 1000.0))

            elif hold_settings and pulse_settings:
                driver.enable(pulse_settings, hold_settings)
                self.drivers_holding_a_side.add(driver)

            else:  # ms == 0
                driver.disable()
                try:
                    self.drivers_holding_a_side.remove(driver)
                except KeyError:
                    pass

    # -------------------------------- C SIDE ---------------------------------

    def _enable_c_side(self):
        if self.prefer_a_side:
            if not self.c_side_enabled and not self.ac_relay_in_transition:

                if self.a_side_active:
                    self._disable_all_a_side_drivers()
                    self._enable_ac_relay()
                    self.delay.add(ms=self.system11_config['ac_relay_delay_ms'],
                                   callback=self._enable_c_side,
                                   name='enable_c_side')
                    return

                if self.a_side_enabled:
                    self._enable_ac_relay()

                else:
                    self._c_side_enabled()
        else:
            if (not self.ac_relay_in_transition and
                    not self.c_side_enabled and
                    not self.a_side_busy):
                self._enable_ac_relay()

            elif self.c_side_enabled and self.c_side_queue:
                self._service_c_side()

    def _c_side_enabled(self):
        self.ac_relay_in_transition = False

        if self.prefer_a_side:
            self.c_side_enabled = True
            self.a_side_enabled = False
            self._service_c_side()
        else:

            if self.a_side_queue:
                self._enable_a_side()
                return

            self.a_side_enabled = False
            self.c_side_enabled = True
            self._service_c_side()

    def _service_c_side(self):
        if not self.c_side_queue:
            return

        if self.ac_relay_in_transition or self.a_side_busy:
            return

        if not self.c_side_enabled:
            self._enable_c_side()
            return

        while self.c_side_queue:
            driver, pulse_settings, hold_settings = self.c_side_queue.pop()

            if hold_settings is None and pulse_settings:
                driver.pulse(pulse_settings)
                self.c_side_done_time = max(self.c_side_done_time,
                                            self.machine.clock.get_time() + (pulse_settings.duration / 1000.0))
            elif hold_settings and pulse_settings:
                driver.enable(pulse_settings, hold_settings)
                self.drivers_holding_c_side.add(driver)

            else:
                driver.disable()
                try:
                    self.drivers_holding_c_side.remove(driver)
                except KeyError:
                    pass

    def _disable_all_c_side_drivers(self):
        if self.c_side_active:
            for driver in self.drivers_holding_c_side:
                driver.disable()
            self.drivers_holding_c_side = set()
            self.c_side_done_time = 0
            self.c_side_enabled = False

    def _disable_all_a_side_drivers(self):
        if self.a_side_active:
            for driver in self.drivers_holding_a_side:
                driver.disable()
            self.drivers_holding_a_side = set()
            self.a_side_done_time = 0
            self.a_side_enabled = False

    def validate_coil_section(self, driver, config):
        """Validate coil config for platform."""
        return self.platform.validate_coil_section(driver, config)


class System11Driver(DriverPlatformInterface):

    """Represent one driver on the system11 overlay.

    Two of those drivers may be created for one real driver. One for the A and one for the C side.
    """

    def __init__(self, number, platform_driver: DriverPlatformInterface, overlay, side) -> None:
        """Initialize driver."""
        super().__init__(platform_driver.config, number)
        self.number = number
        self.platform_driver = platform_driver
        self.overlay = overlay
        self.side = side

    def __repr__(self):
        """Pretty print."""
        return "System11Driver.{}".format(self.number)

    def get_board_name(self):
        """Return name of driver board."""
        return self.platform_driver.get_board_name()

    def pulse(self, pulse_settings: PulseSettings):
        """Pulse driver."""
        self.overlay.driver_action(self.platform_driver, pulse_settings, None, self.side)

        # Usually pulse() returns the value (in ms) that the driver will pulse
        # for so we can update Driver.time_when_done. But with A/C switched
        # coils, we don't know when exactly that will be, so we return -1
        return -1

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Enable driver."""
        self.overlay.driver_action(self.platform_driver, pulse_settings, hold_settings, self.side)

    def disable(self):
        """Disable driver."""
        self.overlay.driver_action(self.platform_driver, None, None, self.side)



from unittest.mock import MagicMock
from mpf.tests.MpfTestCase import MpfTestCase


class TestInfoLights(MpfTestCase):

    def __init__(self, methodName):
        super().__init__(methodName)
        self.machine_config_patches['mpf']['plugins'] = ["mpf.plugins.info_lights.InfoLights"]

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/info_lights/'

    def start_game(self):
        # shots only work in games so we have to do this a lot
        self.machine.playfield.add_ball = MagicMock()
        self.machine.ball_controller.num_balls_known = 3
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertEqual(2, self.machine.game.num_players)
        self.assertIsNotNone(self.machine.game)

    def stop_game(self):
        # stop game
        self.assertIsNotNone(self.machine.game)
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertIsNone(self.machine.game)

    def test_info_lights(self):
        # machine starts at gameover
        self.advance_time_and_run(1)
        self.assertLightColor("gameOver", [0, 0, 0])
        self.advance_time_and_run(1)
        self.assertLightColor("gameOver", [255, 255, 255])
        self.advance_time_and_run(1)
        self.assertLightColor("gameOver", [0, 0, 0])

        self.start_game()

        # no more game over
        self.assertLightColor("gameOver", [0, 0, 0])
        self.advance_time_and_run(1)
        self.assertLightColor("gameOver", [0, 0, 0])

        # ball one
        self.assertLightChannel("bip1", 255)
        self.assertLightChannel("bip2", 0)
        self.assertLightChannel("bip3", 0)
        # no tilt
        self.assertLightColor("tilt", [0, 0, 0])
        # two players
        self.assertLightChannel("player1", 255)
        self.assertLightChannel("player2", 255)

        self.machine.game.balls_in_play = 0
        self.advance_time_and_run()

        # ball one
        self.assertLightChannel("bip1", 255)
        self.assertLightChannel("bip2", 0)
        self.assertLightChannel("bip3", 0)
        # no tils
        self.assertLightColor("tilt", [0, 0, 0])

        self.machine.game.balls_in_play = 0
        self.advance_time_and_run()

        # ball two
        self.assertLightChannel("bip1", 0)
        self.assertLightChannel("bip2", 255)
        self.assertLightChannel("bip3", 0)
        # no tilt
        self.assertLightColor("tilt", [0, 0, 0])

        self.stop_game()
        self.advance_time_and_run(.5)

        # gameover
        self.assertLightColor("gameOver", [0, 0, 0])
        self.advance_time_and_run(1)
        self.assertLightColor("gameOver", [255, 255, 255])
        self.advance_time_and_run(1)
        self.assertLightColor("gameOver", [0, 0, 0])

        self.start_game()
        self.advance_time_and_run()

        self.post_event("tilt")
        self.advance_time_and_run()

        # no ball
        self.assertLightChannel("bip1", 255)
        self.assertLightChannel("bip2", 0)
        self.assertLightChannel("bip3", 0)
        # tilt
        self.assertLightColor("tilt", [255, 255, 255])

        self.assertLightChannel("player1", 255)
        self.assertLightChannel("player2", 255)




from mpf.tests.MpfGameTestCase import MpfGameTestCase


class TestBallRouting(MpfGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_routing/'

    def get_platform(self):
        return 'smart_virtual'

    def test_routing(self):
        self.hit_switch_and_run("s_ball_switch1", 1)
        self.hit_switch_and_run("s_ball_switch2", 1)
        self.start_game()
        self.advance_time_and_run(10)

        self.assertBallsOnPlayfield(1)

        # by default the ball is routed back to the playfield
        self.mock_event("balldevice_test_device1_ball_enter")
        self.mock_event("balldevice_test_device2_ball_enter")
        self.hit_switch_and_run("s_device1", 1)
        self.assertBallsOnPlayfield(0)
        self.assertAvailableBallsOnPlayfield(1)
        self.advance_time_and_run(20)
        self.assertBallsOnPlayfield(1)
        self.assertAvailableBallsOnPlayfield(1)
        self.assertEventCalledWith("balldevice_test_device1_ball_enter",
                                   device=self.machine.ball_devices["test_device1"], new_balls=1, unclaimed_balls=1,
                                   new_available_balls=1)
        self.assertEventCalledWith("balldevice_test_device2_ball_enter",
                                   device=self.machine.ball_devices["test_device2"], new_balls=1, unclaimed_balls=0,
                                   new_available_balls=0)

        self.start_mode("mode1")
        # route ball via device2
        self.post_event("route_to_2")
        self.mock_event("balldevice_test_device1_ball_enter")
        self.mock_event("balldevice_test_device2_ball_enter")
        self.mock_event("balldevice_test_device3_ball_enter")
        self.mock_event("balldevice_test_device4_ball_enter")
        self.hit_switch_and_run("s_device1", 1)
        self.assertBallsOnPlayfield(0)
        self.advance_time_and_run(20)
        self.assertBallsOnPlayfield(1)
        self.assertAvailableBallsOnPlayfield(1)
        self.assertEventCalledWith("balldevice_test_device1_ball_enter",
                                   device=self.machine.ball_devices["test_device1"], new_balls=1, unclaimed_balls=0,
                                   new_available_balls=1)
        self.assertEventCalledWith("balldevice_test_device2_ball_enter",
                                   device=self.machine.ball_devices["test_device2"], new_balls=1, unclaimed_balls=1,
                                   new_available_balls=0)
        self.assertEventNotCalled("balldevice_test_device3_ball_enter")
        self.assertEventNotCalled("balldevice_test_device4_ball_enter")

        # route ball via device4
        self.post_event("route_to_4")
        self.mock_event("balldevice_test_device1_ball_enter")
        self.mock_event("balldevice_test_device2_ball_enter")
        self.mock_event("balldevice_test_device3_ball_enter")
        self.mock_event("balldevice_test_device4_ball_enter")
        self.hit_switch_and_run("s_device1", 1)
        self.assertBallsOnPlayfield(0)
        self.advance_time_and_run(20)
        self.assertBallsOnPlayfield(1)
        self.assertAvailableBallsOnPlayfield(1)
        self.assertEventCalledWith("balldevice_test_device1_ball_enter",
                                   device=self.machine.ball_devices["test_device1"], new_balls=1, unclaimed_balls=0,
                                   new_available_balls=1)
        self.assertEventNotCalled("balldevice_test_device2_ball_enter")
        self.assertEventCalledWith("balldevice_test_device3_ball_enter",
                                   device=self.machine.ball_devices["test_device3"], new_balls=1, unclaimed_balls=0,
                                   new_available_balls=0)
        self.assertEventCalledWith("balldevice_test_device4_ball_enter",
                                   device=self.machine.ball_devices["test_device4"], new_balls=1, unclaimed_balls=1,
                                   new_available_balls=0)
        self.assertEqual(0, self.machine.ball_devices["test_device4"].balls)
        self.assertEqual(0, self.machine.ball_devices["test_device4"].available_balls)

        # route ball to device4 and lock it
        self.post_event("lock_enable")
        self.post_event("route_to_4")
        self.mock_event("balldevice_test_device1_ball_enter")
        self.mock_event("balldevice_test_device2_ball_enter")
        self.mock_event("balldevice_test_device3_ball_enter")
        self.mock_event("balldevice_test_device4_ball_enter")
        self.hit_switch_and_run("s_device1", 1)
        self.assertBallsOnPlayfield(0)
        self.advance_time_and_run(20)
        self.assertBallsOnPlayfield(1)
        self.assertAvailableBallsOnPlayfield(1)
        self.assertEventCalledWith("balldevice_test_device1_ball_enter",
                                   device=self.machine.ball_devices["test_device1"], new_balls=1, unclaimed_balls=0,
                                   new_available_balls=1)
        self.assertEventNotCalled("balldevice_test_device2_ball_enter")
        self.assertEventCalledWith("balldevice_test_device3_ball_enter",
                                   device=self.machine.ball_devices["test_device3"], new_balls=1, unclaimed_balls=0,
                                   new_available_balls=0)
        self.assertEventCalledWith("balldevice_test_device4_ball_enter",
                                   device=self.machine.ball_devices["test_device4"], new_balls=1, unclaimed_balls=0,
                                   new_available_balls=0)
        self.assertEqual(1, self.machine.ball_devices["test_device4"].balls)
        self.assertEqual(1, self.machine.ball_devices["test_device4"].available_balls)



"""MPF plugin which adds events from monitoring a Twitch chat room."""
import threading
from mpf.core.logging import LogMixin

from .twitch.twitch_client import TwitchClient

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class TwitchBot(LogMixin):

    """Adds Twitch chat room events."""

    def __init__(self, machine):
        """Initialise Twitch client."""
        super().__init__()
        self.machine = machine      # type: MachineController
        self.configure_logging("twitch_client")

        if 'twitch_client' not in machine.config:
            self.log.debug('"twitch_client:" section not found in '
                           'machine configuration, so the Twitch Client'
                           'plugin will not be used.')
            return

        self.config = self.machine.config_validator.validate_config(
            "twitch_client", self.machine.config['twitch_client'])

        self.log.info('Attempting to connect to Twitch')

        user_var = self.config['user_var']
        password_var = self.config['password_var']
        channel_var = self.config['channel_var']

        user = self.machine.variables.get_machine_var(
            user_var
        ) if user_var is not None else self.config['user']

        password = self.machine.variables.get_machine_var(
            password_var
        ) if password_var is not None else self.config['password']

        channel = self.machine.variables.get_machine_var(
            channel_var
        ) if channel_var is not None else self.config['channel']

        self.client = TwitchClient(self.machine, user, password, channel,
                                   self.machine.clock.loop)
        thread = threading.Thread(target=self.client.start, args=())
        thread.daemon = True
        thread.start()

        if self.client.is_connected():
            self.info_log('Successful connection to Twitch')
        else:
            self.info_log('Connecting...')



"""RGBA Color."""
from typing import Tuple, Union, List

from mpf.core.rgb_color import RGBColor


class RGBAColor(RGBColor):

    """RGB Color with alpha channel."""

    __slots__ = ["opacity"]

    def __init__(self, color: Union[RGBColor, str, Tuple[int, int, int], Tuple[int, int, int, int], List[int]]) -> None:
        """Initialise RGBA color."""
        if isinstance(color, (tuple, list)) and len(color) == 4:
            self.opacity = color[3]     # type: ignore
            super().__init__((color[0], color[1], color[2]))
        else:
            self.opacity = 255
            super().__init__(color)     # type: ignore

    def __iter__(self):
        """Return iterator."""
        return iter([self._color[0], self._color[1], self._color[2], self.opacity])

    def __str__(self):
        """Return string representation."""
        return "{} Opacity: {}".format(self._color, self.opacity)

    @property
    def rgba(self) -> Tuple[int, int, int, int]:
        """Return an RGB representation of the color."""
        return self._color[0], self._color[1], self._color[2], self.opacity

    @rgba.setter
    def rgba(self, value: Tuple[int, int, int, int]):
        self._color = value[0:3]
        self.opacity = value[3]



from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase

from mpf.tests.MpfTestCase import MpfTestCase, test_config
from unittest.mock import MagicMock


class TestDiverter(MpfTestCase):

    def get_machine_path(self):
        return 'tests/machine_files/diverter/'

    def get_platform(self):
        return "smart_virtual"

    def start_game(self):
        # shots only work in games so we have to do this a lot
        self.machine.events.post('game_start')
        self.advance_time_and_run()
        self.machine.game.balls_in_play = 1
        self.assertIsNotNone(self.machine.game)

    def _block_device(self, queue, **kwargs):
        del kwargs
        self.queue = queue
        queue.wait()

    @test_config("only_events_no_coils.yaml")
    def test_diverter_with_servo_and_stepper(self):
        self.queue = None
        diverter = self.machine.diverters["d_test_with_events"]
        self.advance_time_and_run(.1)

        self.assertEqual(.2, self.machine.servos["s_diverter"].hw_servo.current_position)
        self.assertEqual(400, self.machine.steppers["s_diverter"].hw_stepper._current_position)


        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_trough"]
        self.machine.playfield.add_ball()

        self.advance_time_and_run(1)

        self.assertTrue(diverter.enabled)
        self.assertTrue(diverter.active)
        self.assertEqual(.7, self.machine.servos["s_diverter"].hw_servo.current_position)
        self.assertEqual(20, self.machine.steppers["s_diverter"].hw_stepper._current_position)

        self.hit_and_release_switch("s_playfield")
        self.machine_run()
        self.assertFalse(diverter.active)

        self.assertEqual(.2, self.machine.servos["s_diverter"].hw_servo.current_position)
        self.assertEqual(400, self.machine.steppers["s_diverter"].hw_stepper._current_position)

        self.hit_switch_and_run("s_ball_switch1", 1)
        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_target"]
        self.machine.playfield.add_ball()

        self.advance_time_and_run(3)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.hit_and_release_switch("s_diverter")
        self.advance_time_and_run(0.5)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.assertEqual(.2, self.machine.servos["s_diverter"].hw_servo.current_position)
        self.assertEqual(400, self.machine.steppers["s_diverter"].hw_stepper._current_position)

        self.advance_time_and_run(4)
        self.assertEqual(0, diverter.diverting_ejects_count)

    @test_config("test_delayed_eject.yaml")
    def test_delayed_eject(self):
        self.queue = None
        diverter = self.machine.diverters["d_test_delayed_eject"]

        self.machine.coils["c_diverter"].enable = MagicMock()
        self.machine.coils["c_diverter"].disable = MagicMock()

        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.machine.events.add_handler("balldevice_test_trough_ball_eject_attempt", self._block_device)

        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_trough"]
        self.machine.playfield.add_ball()

        self.advance_time_and_run(20)
        self.queue.clear()

        self.advance_time_and_run(1)
        self.machine.events.remove_handler_by_event("balldevice_test_trough_ball_eject_attempt", self._block_device)

        self.assertTrue(diverter.enabled)
        self.assertTrue(diverter.active)
        self.machine.coils["c_diverter"].enable.assert_called_once_with()
        self.machine.coils["c_diverter"].enable = MagicMock()
        assert not self.machine.coils["c_diverter"].disable.called

        self.advance_time_and_run(4)
        self.machine.coils["c_diverter"].disable.assert_called_once_with()
        assert not self.machine.coils["c_diverter"].enable.called

        self.hit_and_release_switch("s_playfield")
        self.machine_run()
        self.assertFalse(diverter.active)

        self.hit_switch_and_run("s_ball_switch1", 1)
        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_target"]
        self.machine.playfield.add_ball()

        self.advance_time_and_run(3)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.hit_and_release_switch("s_diverter")
        self.advance_time_and_run(0.5)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.advance_time_and_run(4)
        assert not self.machine.coils["c_diverter"].enable.called
        self.assertEqual(0, diverter.diverting_ejects_count)

    @test_config("test_hold_activation_time.yaml")
    def test_hold_activation_time(self):
        diverter = self.machine.diverters["d_test_hold_activation_time"]

        self.machine.coils["c_diverter"].enable = MagicMock()
        self.machine.coils["c_diverter"].disable = MagicMock()

        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_trough"]
        self.machine.playfield.add_ball()

        self.advance_time_and_run(1)
        self.assertTrue(diverter.enabled)
        self.assertFalse(diverter.active)

        self.hit_and_release_switch("s_diverter")
        self.advance_time_and_run(0.5)
        self.assertTrue(diverter.enabled)
        self.assertTrue(diverter.active)
        self.machine.coils["c_diverter"].enable.assert_called_once_with()
        self.machine.coils["c_diverter"].enable = MagicMock()
        assert not self.machine.coils["c_diverter"].disable.called

        self.advance_time_and_run(4)
        self.machine.coils["c_diverter"].disable.assert_called_once_with()
        assert not self.machine.coils["c_diverter"].enable.called

        self.hit_and_release_switch("s_playfield")
        self.machine_run()
        self.assertFalse(diverter.active)

        self.hit_switch_and_run("s_ball_switch1", 1)
        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_target"]
        self.machine.playfield.add_ball()

        self.advance_time_and_run(3)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.hit_and_release_switch("s_diverter")
        self.advance_time_and_run(0.5)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.advance_time_and_run(4)
        assert not self.machine.coils["c_diverter"].enable.called

    @test_config("test_hold_no_activation_time.yaml")
    def test_hold_no_activation_time(self):
        diverter = self.machine.diverters["d_test_hold"]

        self.assertEqual("idle", self.machine.ball_devices["test_trough"]._state)
        self.assertEqual("idle", self.machine.ball_devices["test_target"]._state)

        self.machine.coils["c_diverter"].enable = MagicMock()
        self.machine.coils["c_diverter"].disable = MagicMock()

        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        # test active side
        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_trough"]
        self.machine.playfield.add_ball()

        self.advance_time_and_run(1)
        self.assertTrue(diverter.enabled)
        self.assertFalse(diverter.active)

        self.hit_and_release_switch("s_diverter")
        self.advance_time_and_run(0.5)
        self.assertTrue(diverter.enabled)
        self.assertTrue(diverter.active)
        self.machine.coils["c_diverter"].enable.assert_called_once_with()
        self.machine.coils["c_diverter"].enable = MagicMock()
        assert not self.machine.coils["c_diverter"].disable.called

        self.advance_time_and_run(4)
        assert not self.machine.coils["c_diverter"].disable.called

        self.hit_and_release_switch("s_playfield")
        self.machine_run()
        self.assertFalse(diverter.active)
        self.machine.coils["c_diverter"].disable.assert_called_once_with()

        self.hit_switch_and_run("s_ball_switch1", 1)
        # test inactive side
        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_target"]
        self.machine.playfield.add_ball()

        self.advance_time_and_run(3)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.hit_and_release_switch("s_diverter")
        self.advance_time_and_run(0.5)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.advance_time_and_run(4)
        assert not self.machine.coils["c_diverter"].enable.called

        self.hit_switch_and_run("s_ball_switch1", 1)

    @test_config("test_missing_ball_at_source.yaml")
    def test_missing_ball_at_source(self):
        diverter = self.machine.diverters["d_test"]
        trough1 = self.machine.ball_devices["test_trough"]
        target = self.machine.ball_devices["test_target"]

        self.machine.default_platform.actions[trough1].set_result("missing")
        trough1.eject(1, target)
        self.advance_time_and_run(1)
        self.assertEqual(1, diverter.diverting_ejects_count)

        self.advance_time_and_run(100)

        self.assertEqual(0, diverter.diverting_ejects_count)

    @test_config("test_eject_to_oposide_sides2.yaml")
    def test_eject_to_oposide_sides2(self):
        diverter = self.machine.diverters["d_test_hold"]
        trough1 = self.machine.ball_devices["test_trough"]
        trough2 = self.machine.ball_devices["test_trough2"]
        playfield = self.machine.playfield
        target = self.machine.ball_devices["test_target"]

        self.assertEqual(3, trough1.balls)
        self.assertEqual(3, trough2.balls)

        trough1.eject(1, playfield)
        trough2.eject(1, target)
        trough1.eject(1, playfield)
        trough2.eject(1, playfield)
        trough1.eject(1, target)
        trough2.eject(1, target)

        self.advance_time_and_run(100)

        self.assertEqual("idle", trough1._state)
        self.assertEqual("idle", trough2._state)

        self.assertEqual(0, diverter.diverting_ejects_count)

        self.assertEqual(3, playfield.balls)
        self.assertEqual(3, target.balls)
        self.assertEqual(0, trough1.balls)
        self.assertEqual(0, trough2.balls)

    @test_config("test_eject_to_oposide_sides.yaml")
    def test_eject_to_oposide_sides(self):
        diverter = self.machine.diverters["d_test_hold"]
        trough1 = self.machine.ball_devices["test_trough"]
        trough2 = self.machine.ball_devices["test_trough2"]
        playfield = self.machine.playfield
        target = self.machine.ball_devices["test_target"]

        pulse1 = self.machine.coils["eject_coil1"].pulse
        pulse3 = self.machine.coils["eject_coil3"].pulse

        self.machine.coils["eject_coil1"].pulse = MagicMock(wraps=pulse1)
        self.machine.coils["eject_coil3"].pulse = MagicMock(wraps=pulse3)

        # this goes to the active side of the diverter
        trough1.eject(1, playfield)
        # this goes to the inactive side of the diverter
        trough2.eject(1, target)

        self.advance_time_and_run(5)

        self.assertEqual("ball_left", trough1._state)
        self.assertEqual("waiting_for_target_ready", trough2._state)

        self.assertTrue(diverter.enabled)
        self.assertTrue(diverter.active)

        # only trough1 ejects. trought2 waits because of diverter
        self.assertTrue(self.machine.coils["eject_coil1"].pulse.called)
        assert not self.machine.coils["eject_coil3"].pulse.called
        self.assertEqual(1, diverter.diverting_ejects_count)

        self.hit_and_release_switch("s_playfield")
        self.advance_time_and_run(3)
        # wait for cooldown
        self.assertFalse(self.machine.coils["eject_coil3"].pulse.called)
        self.assertEqual(1, diverter.diverting_ejects_count)
        self.advance_time_and_run(3)
        self.assertTrue(self.machine.coils["eject_coil3"].pulse.called)
        self.assertEqual(1, diverter.diverting_ejects_count)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.machine.coils["eject_coil1"].pulse = MagicMock(wraps=pulse1)
        self.machine.coils["eject_coil3"].pulse = MagicMock(wraps=pulse3)

        self.machine.ball_devices["test_target"].eject()
        self.advance_time_and_run(20)

        self.assertEqual("idle", trough1._state)
        self.assertEqual("idle", trough2._state)
        self.assertEqual(2, trough1.balls)
        self.assertEqual(1, trough2.balls)

        self.assertEqual("idle", target._state)
        self.assertEqual(0, diverter.diverting_ejects_count)

        self.machine.log.info("START")

        # same scenario but change order
        trough2.eject(1, target)
        trough1.eject(1, playfield)

        self.advance_time_and_run(0.05)

        self.assertEqual("waiting_for_target_ready", trough1._state)
        self.assertEqual("ball_left", trough2._state)

        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.assertTrue(self.machine.coils["eject_coil3"].pulse.called)
        assert not self.machine.coils["eject_coil1"].pulse.called

        self.hit_and_release_switch("s_playfield")
        self.advance_time_and_run(3)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        # wait for cooldown
        self.advance_time_and_run(3)

        self.assertTrue(diverter.enabled)
        self.assertTrue(diverter.active)

        self.assertTrue(self.machine.coils["eject_coil1"].pulse.called)
        self.advance_time_and_run(20)

        self.assertEqual(0, diverter.diverting_ejects_count)

    @test_config("test_pulsed_activation_time.yaml")
    def test_pulsed_activation_time(self):
        diverter = self.machine.diverters["d_test_pulse"]

        self.machine.coils["c_diverter"].pulse = MagicMock()
        self.machine.coils["c_diverter_disable"].pulse = MagicMock()

        self.post_event("machine_reset_phase_3")
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)
        self.assertTrue(self.machine.coils["c_diverter_disable"].pulse.called)
        assert not self.machine.coils["c_diverter"].pulse.called

        self.assertEqual("idle", self.machine.ball_devices["test_trough"]._state)
        self.assertEqual("idle", self.machine.ball_devices["test_target"]._state)

        self.machine.coils["c_diverter"].pulse = MagicMock()
        self.machine.coils["c_diverter_disable"].pulse = MagicMock()

        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_trough"]
        self.machine.playfield.add_ball()

        self.advance_time_and_run(1)
        self.assertTrue(diverter.enabled)
        self.assertTrue(diverter.active)
        self.assertTrue(self.machine.coils["c_diverter"].pulse.called)
        self.machine.coils["c_diverter"].pulse = MagicMock()
        assert not self.machine.coils["c_diverter_disable"].pulse.called

        self.advance_time_and_run(4)
        assert not self.machine.coils["c_diverter_disable"].pulse.called

        self.hit_and_release_switch("s_playfield")
        self.machine_run()
        self.assertFalse(diverter.active)
        self.assertTrue(self.machine.coils["c_diverter_disable"].pulse.called)

        self.hit_switch_and_run("s_ball_switch1", 1)
        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_target"]
        self.machine.playfield.add_ball()

        self.advance_time_and_run(3)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.hit_and_release_switch("s_diverter")
        self.advance_time_and_run(0.5)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.advance_time_and_run(4)
        assert not self.machine.coils["c_diverter"].pulse.called
        self.assertEqual(0, diverter.diverting_ejects_count)

    @test_config("test_diverter_with_switch.yaml")
    def test_diverter_with_switch(self):
        diverter = self.machine.diverters["d_test"]

        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        # it should not activate
        self.hit_and_release_switch("s_activate")
        self.advance_time_and_run(1)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        # nothing should happen
        self.hit_and_release_switch("s_deactivate")
        self.hit_and_release_switch("s_disable")
        self.advance_time_and_run(1)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        # enable diverter
        diverter.enable()
        self.assertTrue(diverter.enabled)
        self.assertFalse(diverter.active)

        self.hit_and_release_switch("s_activate")
        self.advance_time_and_run(1)
        self.assertTrue(diverter.enabled)
        self.assertTrue(diverter.active)

        self.hit_and_release_switch("s_deactivate")
        self.advance_time_and_run(1)
        self.assertTrue(diverter.enabled)
        self.assertFalse(diverter.active)

        self.hit_and_release_switch("s_activate")
        self.advance_time_and_run(1)
        self.assertTrue(diverter.enabled)
        self.assertTrue(diverter.active)

        self.hit_and_release_switch("s_disable")
        self.advance_time_and_run(1)
        self.assertFalse(diverter.enabled)
        self.assertTrue(diverter.active)

        self.hit_and_release_switch("s_deactivate")
        self.advance_time_and_run(1)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

    @test_config("test_diverter_auto_disable.yaml")
    def test_diverter_auto_disable(self):
        diverter = self.machine.diverters["d_test"]

        # enable diverter
        diverter.enable()
        self.assertTrue(diverter.enabled)
        self.assertFalse(diverter.active)

        self.hit_and_release_switch("s_activate")
        self.advance_time_and_run(1)
        self.assertTrue(diverter.enabled)
        self.assertTrue(diverter.active)

        self.hit_and_release_switch("s_disable")
        self.advance_time_and_run(1)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

    @test_config("test_activation_switch_and_eject_confirm_switch.yaml")
    def test_activation_switch_and_eject_confirm_switch(self):
        diverter = self.machine.diverters["d_test_hold_activation_time"]

        self.machine.coils["c_diverter"].enable = MagicMock()
        self.machine.coils["c_diverter"].disable = MagicMock()

        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_trough2"]
        self.machine.playfield.add_ball()

        self.advance_time_and_run(.01)
        self.assertTrue(diverter.enabled)
        self.assertFalse(diverter.active)

        # smart virtual automatically triggers the diverter
        self.advance_time_and_run(3)
        self.assertTrue(diverter.enabled)
        self.assertTrue(diverter.active)
        self.machine.coils["c_diverter"].enable.assert_called_once_with()
        self.machine.coils["c_diverter"].enable = MagicMock()
        assert not self.machine.coils["c_diverter"].disable.called

        self.advance_time_and_run(4)
        self.machine.coils["c_diverter"].disable.assert_called_once_with()
        assert not self.machine.coils["c_diverter"].enable.called

        self.hit_and_release_switch("s_playfield")
        self.machine_run()
        self.assertFalse(diverter.active)

        self.hit_switch_and_run("s_ball_switch1", 1)
        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_target"]
        self.machine.ball_devices["test_trough2"].eject(1, self.machine.ball_devices["test_target"])

        self.advance_time_and_run(.01)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        # smart virtual htis diverter switch for us
        self.advance_time_and_run(3)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.advance_time_and_run(4)
        assert not self.machine.coils["c_diverter"].enable.called
        self.assertEqual(0, diverter.diverting_ejects_count)

    @test_config("test_diverter_dual_wound_coil.yaml")
    def test_diverter_dual_wound_coil(self):
        diverter = self.machine.diverters["d_test_dual_wound"]

        self.assertEqual("idle", self.machine.ball_devices["test_trough"]._state)
        self.assertEqual("idle", self.machine.ball_devices["test_target"]._state)

        self.machine.coils["c_hold"].enable = MagicMock()
        self.machine.coils["c_hold"].disable = MagicMock()
        self.machine.coils["c_power"].pulse = MagicMock()

        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        # test active side
        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_trough"]
        self.machine.playfield.add_ball()

        self.advance_time_and_run(1)
        self.assertTrue(diverter.enabled)
        self.assertFalse(diverter.active)

        self.hit_and_release_switch("s_diverter")
        self.advance_time_and_run(0.5)
        self.assertTrue(diverter.enabled)
        self.assertTrue(diverter.active)
        self.machine.coils["c_hold"].enable.assert_called_once_with()
        self.machine.coils["c_hold"].enable = MagicMock()
        self.assertTrue(self.machine.coils["c_power"].pulse.called)
        self.machine.coils["c_power"].pulse = MagicMock()
        assert not self.machine.coils["c_hold"].disable.called

        self.advance_time_and_run(4)
        assert not self.machine.coils["c_hold"].disable.called

        self.hit_and_release_switch("s_playfield")
        self.machine_run()
        self.assertFalse(diverter.active)
        self.machine.coils["c_hold"].disable.assert_called_once_with()

        self.hit_switch_and_run("s_ball_switch1", 1)
        # test inactive side
        self.machine.playfield.config['default_source_device'] = self.machine.ball_devices["test_target"]
        self.machine.playfield.add_ball()

        self.advance_time_and_run(3)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.hit_and_release_switch("s_diverter")
        self.advance_time_and_run(0.5)
        self.assertFalse(diverter.enabled)
        self.assertFalse(diverter.active)

        self.advance_time_and_run(4)
        assert not self.machine.coils["c_hold"].enable.called
        assert not self.machine.coils["c_power"].pulse.called

        self.hit_switch_and_run("s_ball_switch1", 1)


class TestDiverter2(MpfFakeGameTestCase):


    def get_machine_path(self):
        return 'tests/machine_files/diverter/'

    @test_config("diverter_with_activation_events.yaml")
    def testActivationEvents(self):
        """Make sure diverter does not activate on enable when activation_events are present."""
        self.mock_event("diverter_test_diverter_enabling")
        self.mock_event("diverter_test_diverter_activating")

        # start game
        self.start_game()
        # diverter should enable but not activate
        self.assertEventCalled("diverter_test_diverter_enabling")
        self.assertEventNotCalled("diverter_test_diverter_activating")

        # activate diverter
        self.post_event("activate_test_diverter")
        # it should activate
        self.assertEventCalled("diverter_test_diverter_activating")



"""Platform interface for smart steppers."""
from typing import List

import abc


class StepperPlatformInterface(metaclass=abc.ABCMeta):

    """Interface for smart steppers in hardware platforms.

    Smart steppers is being used here to designate axis controllers that add motion control over the top of a regular
    direction and step type interface. Including homing, positioning, and velocity modes.

    StepperPlatformInterface is an abstract base class that should be overridden for all
    smart stepper/axis interface classes on supported platforms.  This class ensures the proper required
    methods are implemented to support smart stepper operations in MPF.
    """

    __slots__ = []  # type: List[str]

    @abc.abstractmethod
    def home(self, direction):
        """Home an axis, resetting 0 position.

        direction can be clockwise or counterclockwise.
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def wait_for_move_completed(self):
        """Return after the last move completed.

        This is also used to check if homing is completed.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def move_rel_pos(self, position):
        """Move axis to a certain relative position."""
        raise NotImplementedError

    @abc.abstractmethod
    def move_vel_mode(self, velocity):
        """Move at a specific velocity (pos = clockwise, neg = counterclockwise)."""
        raise NotImplementedError

    @abc.abstractmethod
    def stop(self):
        """Stop the stepper."""
        raise NotImplementedError

    def set_home_position(self):
        """Tell the hardware that it reached the home position."""



"""Contains the Driver parent class."""
from typing import Optional

from mpf.core.delays import DelayManager
from mpf.core.events import event_handler
from mpf.core.machine import MachineController
from mpf.core.platform import DriverPlatform, DriverConfig
from mpf.core.system_wide_device import SystemWideDevice
from mpf.exceptions.driver_limits_error import DriverLimitsError
from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings


class Driver(SystemWideDevice):

    """Generic class that holds driver objects.

    A 'driver' is any device controlled from a driver board which is typically
    the high-voltage stuff like coils and flashers.

    This class exposes the methods you should use on these driver types of
    devices. Each platform module (i.e. P-ROC, FAST, etc.) subclasses this
    class to actually communicate with the physical hardware and perform the
    actions.

    Args: Same as the Device parent class
    """

    config_section = 'coils'
    collection = 'coils'
    class_label = 'coil'

    __slots__ = ["hw_driver", "delay", "platform", "__dict__"]

    def __init__(self, machine: MachineController, name: str) -> None:
        """Initialise driver."""
        self.hw_driver = None   # type: Optional[DriverPlatformInterface]
        super().__init__(machine, name)
        self.delay = DelayManager(self.machine)
        self.platform = None                # type: Optional[DriverPlatform]

    @classmethod
    def device_class_init(cls, machine: MachineController):
        """Register handler for duplicate coil number checks."""
        machine.events.add_handler("init_phase_4", cls._check_duplicate_coil_numbers, machine=machine)

    @staticmethod
    def _check_duplicate_coil_numbers(machine, **kwargs):
        del kwargs
        check_set = set()
        for coil in machine.coils.values():
            if not hasattr(coil, "hw_driver"):
                # skip dual wound and other special devices
                continue
            key = (coil.config['platform'], coil.hw_driver.number)
            if key in check_set:
                raise AssertionError("Duplicate coil number {} for coil {}".format(coil.hw_driver.number, coil))

            check_set.add(key)

    def validate_and_parse_config(self, config: dict, is_mode_config: bool, debug_prefix: str = None) -> dict:
        """Return the parsed and validated config.

        Args:
        ----
            config: Config of device
            is_mode_config: Whether this device is loaded in a mode or system-wide
            debug_prefix: Prefix to use when logging.

        Returns: Validated config
        """
        config = super().validate_and_parse_config(config, is_mode_config, debug_prefix)
        platform = self.machine.get_platform_sections('coils', getattr(config, "platform", None))
        platform.assert_has_feature("drivers")
        config['platform_settings'] = platform.validate_coil_section(self, config.get('platform_settings', None))
        return config

    async def _initialize(self):
        await super()._initialize()
        self.platform = self.machine.get_platform_sections('coils', self.config['platform'])

        config = DriverConfig(
            name=self.name,
            default_pulse_ms=self.get_and_verify_pulse_ms(None),
            default_pulse_power=self.get_and_verify_pulse_power(None),
            default_hold_power=self.get_and_verify_hold_power(None),
            default_recycle=self.config['default_recycle'],
            max_pulse_ms=self.config['max_pulse_ms'],
            max_pulse_power=self.config['max_pulse_power'],
            max_hold_power=self.config['max_hold_power'])
        platform_settings = dict(self.config['platform_settings']) if self.config['platform_settings'] else dict()

        if not self.platform.features['allow_empty_numbers'] and self.config['number'] is None:
            self.raise_config_error("Driver must have a number.", 1)

        try:
            self.hw_driver = self.platform.configure_driver(config, self.config['number'], platform_settings)
        except AssertionError as e:
            raise AssertionError("Failed to configure driver {} in platform. See error above".format(self.name)) from e

    def get_and_verify_pulse_power(self, pulse_power: Optional[float]) -> float:
        """Return the pulse power to use.

        If pulse_power is None it will use the default_pulse_power. Additionally it will verify the limits.
        """
        if pulse_power is None:
            pulse_power = self.config['default_pulse_power'] if self.config['default_pulse_power'] is not None else 1.0

        if pulse_power and 0 > pulse_power > 1:
            raise AssertionError("Pulse power has to be between 0 and 1 but is {}".format(pulse_power))

        max_pulse_power = 0
        if self.config['max_pulse_power']:
            max_pulse_power = self.config['max_pulse_power']
        elif self.config['default_pulse_power']:
            max_pulse_power = self.config['default_pulse_power']

        if pulse_power > max_pulse_power:
            raise DriverLimitsError("Driver may {} not be pulsed with pulse_power {} because max_pulse_power is {}".
                                    format(self.name, pulse_power, max_pulse_power))
        return pulse_power

    def get_and_verify_hold_power(self, hold_power: Optional[float]) -> float:
        """Return the hold power to use.

        If hold_power is None it will use the default_hold_power. Additionally it will verify the limits.
        """
        if hold_power is None and self.config['default_hold_power']:
            hold_power = self.config['default_hold_power']

        if hold_power is None and self.config['max_hold_power']:
            hold_power = self.config['max_hold_power']

        if hold_power is None and self.config['allow_enable']:
            hold_power = 1.0

        if hold_power is None:
            hold_power = 0.0

        if hold_power and 0 > hold_power > 1:
            raise AssertionError("Hold_power has to be between 0 and 1 but is {}".format(hold_power))

        max_hold_power = 0      # type: float
        if self.config['max_hold_power']:
            max_hold_power = self.config['max_hold_power']
        elif self.config['allow_enable']:
            max_hold_power = 1.0
        elif self.config['default_hold_power']:
            max_hold_power = self.config['default_hold_power']

        if hold_power > max_hold_power:
            raise DriverLimitsError("Driver {} may not be enabled with hold_power {} because max_hold_power is {}".
                                    format(self.name, hold_power, max_hold_power))
        return hold_power

    def get_and_verify_pulse_ms(self, pulse_ms: Optional[int]) -> int:
        """Return and verify pulse_ms to use.

        If pulse_ms is None return the default.
        """
        assert self.platform is not None
        if pulse_ms is None:
            if self.config['default_pulse_ms'] is not None:
                pulse_ms = self.config['default_pulse_ms']
            else:
                pulse_ms = self.machine.config['mpf']['default_pulse_ms']

        if not isinstance(pulse_ms, int):
            raise AssertionError("Wrong type {}".format(pulse_ms))

        if 0 > pulse_ms > self.platform.features['max_pulse']:
            raise AssertionError("Pulse_ms {} is not valid.".format(pulse_ms))

        if self.config['max_pulse_ms'] and pulse_ms > self.config['max_pulse_ms']:
            raise DriverLimitsError("Driver {} may not be pulsed with pulse_ms {} because max_pulse_ms is {}".
                                    format(self.name, pulse_ms, self.config['max_pulse_ms']))

        return pulse_ms

    @event_handler(2)
    def event_enable(self, pulse_ms: int = None, pulse_power: float = None, hold_power: float = None, **kwargs):
        """Event handler for control enable."""
        del kwargs
        self.enable(pulse_ms, pulse_power, hold_power)

    def enable(self, pulse_ms: int = None, pulse_power: float = None, hold_power: float = None,
               max_wait_ms: int = None):
        """Enable a driver by holding it 'on'.

        Args:
        ----
            pulse_ms: The number of milliseconds the driver should be
                enabled for. If no value is provided, the driver will be
                enabled for the value specified in the config dictionary.
            pulse_power: The pulse power. A float between 0.0 and 1.0.
            hold_power: The pulse power. A float between 0.0 and 1.0.
            max_wait_ms: Maximum time this pulse may be delayed for PSU optimization.

        If this driver is configured with a holdpatter, then this method will use
        that holdpatter to pwm pulse the driver.

        If not, then this method will just enable the driver. As a safety
        precaution, if you want to enable() this driver without pwm, then you
        have to add the following option to this driver in your machine
        configuration files:

        allow_enable: True
        """
        assert self.hw_driver is not None
        pulse_ms = self.get_and_verify_pulse_ms(pulse_ms)
        wait_ms = self._notify_psu_and_get_wait_ms(pulse_ms, max_wait_ms)

        pulse_power = self.get_and_verify_pulse_power(pulse_power)
        hold_power = self.get_and_verify_hold_power(hold_power)

        if hold_power == 0.0:
            raise DriverLimitsError("Cannot enable driver with hold_power 0.0")

        if wait_ms > 0:
            self.debug_log("Delaying enable by %sms pulse_ms: %sms (%s pulse_power %s hold_power)",
                           wait_ms, pulse_ms, pulse_power, hold_power)
            self.delay.add(wait_ms, self._enable_now, pulse_ms=pulse_ms, pulse_power=pulse_power, hold_power=hold_power)
        else:
            self._enable_now(pulse_ms, pulse_power, hold_power)

    def _enable_now(self, pulse_ms: int = None, pulse_power: float = None, hold_power: float = None):
        self.info_log("Enabling Driver with power %s (pulse_ms %sms and pulse_power %s)", hold_power, pulse_ms,
                      pulse_power)
        self.hw_driver.enable(PulseSettings(power=pulse_power, duration=pulse_ms),
                              HoldSettings(power=hold_power))

        if self.config['max_hold_duration']:
            self.delay.add_if_doesnt_exist(self.config['max_hold_duration'] * 1000, self._enable_limit_reached,
                                           "enable_limit_reached")

        # inform bcp clients
        self.machine.bcp.interface.send_driver_event(action="enable", name=self.name, number=self.config['number'],
                                                     pulse_ms=pulse_ms, pulse_power=pulse_power, hold_power=hold_power)

    def _enable_limit_reached(self):
        """Disable driver and report service alert if max_hold_duration has been reached."""
        self.log.warning("Reached max_hold_duration for this coil. Will disable driver now to prevent damage!")
        self.disable()
        self.machine.service.add_technical_alert(self, "Reached max_hold_duration. Driver disabled to prevent damage!")

    @event_handler(1)
    def event_disable(self, **kwargs):
        """Event handler for disable control event."""
        del kwargs
        self.disable()

    def disable(self):
        """Disable this driver."""
        self.info_log("Disabling Driver")
        self.hw_driver.disable()
        self.delay.remove("enable_limit_reached")
        # inform bcp clients
        self.machine.bcp.interface.send_driver_event(action="disable", name=self.name, number=self.config['number'])

    def _notify_psu_and_get_wait_ms(self, pulse_ms: int, max_wait_ms: Optional[int]) -> int:
        """Determine if this pulse should be delayed."""
        if max_wait_ms is None:
            self.config['psu'].notify_about_instant_pulse(pulse_ms=pulse_ms)
            return 0

        return self.config['psu'].get_wait_time_for_pulse(pulse_ms=pulse_ms, max_wait_ms=max_wait_ms)

    def _pulse_now(self, pulse_ms: int, pulse_power: float) -> None:
        """Pulse this driver now."""
        assert self.hw_driver is not None
        assert self.platform is not None
        if 0 < pulse_ms <= self.platform.features['max_pulse']:
            self.info_log("Pulsing Driver for %sms (%s pulse_power)", pulse_ms, pulse_power)
            self.hw_driver.pulse(PulseSettings(power=pulse_power, duration=pulse_ms))
        else:
            self.info_log("Enabling Driver for %sms (%s pulse_power)", pulse_ms, pulse_power)
            self.delay.reset(name='timed_disable',
                             ms=pulse_ms,
                             callback=self.disable)
            self.hw_driver.enable(PulseSettings(power=pulse_power, duration=0),
                                  HoldSettings(power=pulse_power))
        # inform bcp clients
        self.machine.bcp.interface.send_driver_event(action="pulse", name=self.name, number=self.config['number'],
                                                     pulse_ms=pulse_ms, pulse_power=pulse_power)

    @event_handler(3)
    def event_pulse(self, pulse_ms: int = None, pulse_power: float = None, max_wait_ms: int = None, **kwargs) -> None:
        """Event handler for pulse control events."""
        del kwargs
        self.pulse(pulse_ms, pulse_power, max_wait_ms)

    def pulse(self, pulse_ms: int = None, pulse_power: float = None, max_wait_ms: int = None) -> int:
        """Pulse this driver.

        Args:
        ----
            pulse_ms: The number of milliseconds the driver should be
                enabled for. If no value is provided, the driver will be
                enabled for the value specified in the config dictionary.
            pulse_power: The pulse power. A float between 0.0 and 1.0.
            max_wait_ms: Maximum time this pulse may be delayed for PSU optimization.
        """
        pulse_ms = self.get_and_verify_pulse_ms(pulse_ms)
        pulse_power = self.get_and_verify_pulse_power(pulse_power)
        wait_ms = self._notify_psu_and_get_wait_ms(pulse_ms, max_wait_ms)

        if wait_ms > 0:
            self.debug_log("Delaying pulse by %sms pulse_ms: %sms (%s pulse_power)", wait_ms, pulse_ms, pulse_power)
            self.delay.add(wait_ms, self._pulse_now, pulse_ms=pulse_ms, pulse_power=pulse_power)
        else:
            self._pulse_now(pulse_ms, pulse_power)

        return wait_ms



"""Generate python files from protobufs."""
import glob
import re

from grpc_tools import protoc   # pylint: disable-msg=import-error


protoc.main([
    'grpc_tools.protoc',
    '--proto_path=protobuf/',
    '--python_out=.',
    '--grpc_python_out=.'
] + list(glob.iglob('./protobuf/*.proto')))

# Make pb2 imports in generated scripts relative
for script in glob.iglob('./*_pb2*.py'):
    with open(script, 'r+') as file:
        code = file.read()
        file.seek(0)
        file.write(re.sub(r'\n(import .+_pb2.*)', '\nfrom . \\1', code))
        file.truncate()



import asyncio

from mpf.tests.MpfTestCase import MpfTestCase, MagicMock


class TestSpiBigBang(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/spi_bit_bang/'

    def get_platform(self):
        # no force platform
        return False

    def test_spi_bit_bang(self):

        # wait for o_cs low
        for i in range(100):
            if self.machine.digital_outputs["o_cs"].hw_driver.state == "enabled":
                break
            self.advance_time_and_run(.025)
        else:
            self.fail("Did not find o_cs high")
        self.machine.digital_outputs["o_clock"].hw_driver.pulse = MagicMock()
        # set bit 7
        self.hit_switch_and_run("s_miso", .05)
        self.assertEqual(1, self.machine.digital_outputs["o_clock"].hw_driver.pulse.call_count)
        # unset bit 6
        self.release_switch_and_run("s_miso", .05)
        self.assertEqual(2, self.machine.digital_outputs["o_clock"].hw_driver.pulse.call_count)
        # unset bit 5
        self.release_switch_and_run("s_miso", .05)
        self.assertEqual(3, self.machine.digital_outputs["o_clock"].hw_driver.pulse.call_count)
        # set bit 4
        self.hit_switch_and_run("s_miso", .05)
        self.assertEqual(4, self.machine.digital_outputs["o_clock"].hw_driver.pulse.call_count)
        # set bit 3
        self.hit_switch_and_run("s_miso", .05)
        self.assertEqual(5, self.machine.digital_outputs["o_clock"].hw_driver.pulse.call_count)
        # set bit 2
        self.hit_switch_and_run("s_miso", .05)
        self.assertEqual(6, self.machine.digital_outputs["o_clock"].hw_driver.pulse.call_count)
        # unset bit 1
        self.release_switch_and_run("s_miso", .05)
        self.assertEqual(7, self.machine.digital_outputs["o_clock"].hw_driver.pulse.call_count)
        # set bit 0
        self.hit_switch_and_run("s_miso", .05)
        self.assertEqual(8, self.machine.digital_outputs["o_clock"].hw_driver.pulse.call_count)
        self.advance_time_and_run(.1)

        self.assertSwitchState("s_trough_0", True)
        self.assertSwitchState("s_trough_1", False)
        self.assertSwitchState("s_trough_2", True)
        self.assertSwitchState("s_trough_3", True)
        self.assertSwitchState("s_trough_4", True)
        self.assertSwitchState("s_trough_5", False)
        self.assertSwitchState("s_trough_6", False)
        self.assertSwitchState("s_trough_7", True)




"""Contains the base class for flippers."""
from typing import List
from typing import Optional

from mpf.core.events import event_handler
from mpf.core.platform_controller import HardwareRule, EosRuleSettings

from mpf.core.device_monitor import DeviceMonitor
from mpf.core.platform_controller import SwitchRuleSettings, DriverRuleSettings, PulseRuleSettings, HoldRuleSettings

from mpf.core.system_wide_device import SystemWideDevice


@DeviceMonitor(_enabled="enabled")
class Flipper(SystemWideDevice):

    """Represents a flipper in a pinball machine. Subclass of Device.

    Contains several methods for actions that can be performed on this flipper,
    like :meth:`enable`, :meth:`disable`, etc.

    Flippers have several options, including player buttons, EOS swtiches,
    multiple coil options (pulsing, hold coils, etc.)

    Args:
    ----
        machine: A reference to the machine controller instance.
        name: A string of the name you'll refer to this flipper object as.
    """

    __slots__ = ["_enabled", "_active_rules", "_sw_flipped"]

    config_section = 'flippers'
    collection = 'flippers'
    class_label = 'flipper'

    def __init__(self, machine, name):
        """Initialise flipper."""
        super().__init__(machine, name)

        self._enabled = False
        self._active_rules = []     # type: List[HardwareRule]
        self._sw_flipped = False

    async def _initialize(self):
        await super()._initialize()
        if self.config['include_in_ball_search']:
            self.config['playfield'].ball_search.register(
                self.config['ball_search_order'], self._ball_search, self.name)

    @event_handler(1)
    def event_enable(self, **kwargs):
        """Handle enable control event.

        To prevent multiple rules at the same time we prioritize disable > enable.
        """
        del kwargs
        self.enable()

    def enable(self):
        """Enable the flipper by writing the necessary hardware rules to the hardware controller.

        The hardware rules for coils can be kind of complex given all the
        options, so we've mapped all the options out here. We literally have
        methods to enable the various rules based on the rule letters here,
        which we've implemented below. Keeps it easy to understand. :)

        Note there's a platform feature saved at:
        self.machine.config['platform']['hw_enable_auto_disable']. If True, it
        means that the platform hardware rules will automatically disable a coil
        that has been enabled when the trigger switch is disabled. If False, it
        means the hardware platform needs its own rule to disable the coil when
        the switch is disabled. Methods F and G below check for that feature
        setting and will not be applied to the hardware if it's True.

        Two coils, using EOS switch to indicate the end of the power stroke:
        Rule  Type     Coil  Switch  Action
        A.    Enable   Main  Button  active
        D.    Enable   Hold  Button  active
        E.    Disable  Main  EOS     active

        One coil, using EOS switch:
        Rule  Type     Coil  Switch  Action
        A.    Enable   Main  Button  active
        H.    PWM      Main  EOS     active

        Two coils, not using EOS switch:
        Rule  Type     Coil  Switch  Action
        B.    Pulse    Main  Button  active
        D.    Enable   Hold  Button  active

        One coil, not using EOS switch:
        Rule  Type       Coil  Switch  Action
        C.    Pulse/PWM  Main  button  active

        Use EOS switch for safety (for platforms that support multiple switch
        rules). Note that this rule is the letter "i", not a numeral 1.
        I. Enable power if button is active and EOS is not active
        """
        # prevent duplicate enable
        if self._enabled:
            return

        self._enabled = True

        self.debug_log('Enabling flipper with config: %s', self.config)

        # Apply the proper hardware rules for our config

        if self.config['activation_switch']:
            # only add rules if we are using a switch
            if self.config['use_eos']:
                self._enable_main_coil_eos_cutoff_rule()
            elif self.config['hold_coil']:
                self._enable_main_coil_pulse_rule()
            else:
                self._enable_single_coil_rule()

            if self.config['hold_coil']:
                self._enable_hold_coil_rule()

    @event_handler(10)
    def event_disable(self, **kwargs):
        """Handle disable control event.

        To prevent multiple rules at the same time we prioritize disable > enable.
        """
        del kwargs
        self.disable()

    def disable(self):
        """Disable the flipper.

        This method makes it so the cabinet flipper buttons no longer control
        the flippers. Used when no game is active and when the player has
        tilted.
        """
        if not self._enabled:
            return

        self.debug_log("Disabling")
        for rule in self._active_rules:
            # disable all rules
            self.machine.platform_controller.clear_hw_rule(rule)

        if self._sw_flipped:
            # disable the coils if activated via sw_flip
            self.sw_release()

        self._active_rules = []

        self._enabled = False

    def _get_pulse_ms(self) -> Optional[int]:
        """Return pulse_ms."""
        pulse_ms = self.config['main_coil_overwrite'].get("pulse_ms", None)
        if self.config['power_setting_name']:
            settings_factor = self.machine.settings.get_setting_value(self.config['power_setting_name'])
            if not pulse_ms:
                pulse_ms = self.machine.config['mpf']['default_pulse_ms']
            return int(pulse_ms * settings_factor)

        return pulse_ms

    def _get_hold_pulse_ms(self) -> Optional[int]:
        """Return pulse_ms for hold coil."""
        pulse_ms = self.config['hold_coil_overwrite'].get("pulse_ms", None)
        if self.config['power_setting_name']:
            settings_factor = self.machine.settings.get_setting_value(self.config['power_setting_name'])
            if not pulse_ms:
                pulse_ms = self.machine.config['mpf']['default_pulse_ms']
            return int(pulse_ms * settings_factor)

        return pulse_ms

    def _get_pulse_power(self) -> Optional[float]:
        """Return pulse_power."""
        pulse_power = self.config['main_coil_overwrite'].get("pulse_power", None)
        return pulse_power

    def _get_hold_pulse_power(self) -> Optional[float]:
        """Return pulse_power for hold coil."""
        pulse_power = self.config['hold_coil_overwrite'].get("pulse_power", None)
        return pulse_power

    def _get_hold_power(self) -> Optional[float]:
        """Return hold_power."""
        hold_power = self.config['main_coil_overwrite'].get("hold_power", None)
        return hold_power

    def _enable_single_coil_rule(self):
        self.debug_log('Enabling single coil rule')

        rule = self.machine.platform_controller.set_pulse_on_hit_and_enable_and_release_rule(
            SwitchRuleSettings(switch=self.config['activation_switch'], debounce=False, invert=False),
            DriverRuleSettings(driver=self.config['main_coil'], recycle=False),
            PulseRuleSettings(duration=self._get_pulse_ms(), power=self._get_pulse_power()),
            HoldRuleSettings(power=self._get_hold_power())
        )
        self._active_rules.append(rule)

    def _enable_main_coil_pulse_rule(self):
        self.debug_log('Enabling main coil pulse rule')

        rule = self.machine.platform_controller.set_pulse_on_hit_and_release_rule(
            SwitchRuleSettings(switch=self.config['activation_switch'], debounce=False, invert=False),
            DriverRuleSettings(driver=self.config['main_coil'], recycle=False),
            PulseRuleSettings(duration=self._get_pulse_ms(), power=self._get_pulse_power())
        )
        self._active_rules.append(rule)

    def _enable_hold_coil_rule(self):
        self.debug_log('Enabling hold coil rule')

        rule = self.machine.platform_controller.set_pulse_on_hit_and_enable_and_release_rule(
            SwitchRuleSettings(switch=self.config['activation_switch'], debounce=False, invert=False),
            DriverRuleSettings(driver=self.config['hold_coil'], recycle=False),
            PulseRuleSettings(duration=self._get_hold_pulse_ms(), power=self._get_hold_pulse_power()),
            HoldRuleSettings(power=self._get_hold_power())
        )
        self._active_rules.append(rule)

    def _enable_main_coil_eos_cutoff_rule(self):
        if self.config['hold_coil']:
            self.debug_log('Enabling main coil EOS cutoff rule w/o hold')
            rule = self.machine.platform_controller.set_pulse_on_hit_and_release_and_disable_rule(
                SwitchRuleSettings(switch=self.config['activation_switch'], debounce=False, invert=False),
                SwitchRuleSettings(switch=self.config['eos_switch'], debounce=False, invert=False),
                DriverRuleSettings(driver=self.config['main_coil'], recycle=False),
                PulseRuleSettings(duration=self._get_hold_pulse_ms(), power=self._get_hold_pulse_power()),
                EosRuleSettings(enable_repulse=self.config["repulse_on_eos_open"],
                                debounce_ms=self.config["eos_active_ms_before_repulse"])
            )
            self._active_rules.append(rule)
        else:
            self.debug_log('Enabling main coil EOS cutoff rule w/ hold')
            rule = self.machine.platform_controller.set_pulse_on_hit_and_enable_and_release_and_disable_rule(
                SwitchRuleSettings(switch=self.config['activation_switch'], debounce=False, invert=False),
                SwitchRuleSettings(switch=self.config['eos_switch'], debounce=False, invert=False),
                DriverRuleSettings(driver=self.config['main_coil'], recycle=False),
                PulseRuleSettings(duration=self._get_hold_pulse_ms(), power=self._get_hold_pulse_power()),
                HoldRuleSettings(power=self._get_hold_power()),
                EosRuleSettings(enable_repulse=self.config["repulse_on_eos_open"],
                                debounce_ms=self.config["eos_active_ms_before_repulse"])
            )
            self._active_rules.append(rule)

    @event_handler(6)
    def event_sw_flip(self, **kwargs):
        """Handle sw_flip control event."""
        del kwargs
        self.sw_flip()

    def sw_flip(self):
        """Activate the flipper via software as if the flipper button was pushed.

        This is needed because the real flipper activations are handled in
        hardware, so if you want to flip the flippers with the keyboard or OSC
        interfaces, you have to call this method.

        Note this method will keep this flipper enabled until you call
        sw_release().
        """
        if not self._enabled:
            return

        self._sw_flipped = True

        if self.config['hold_coil']:
            self.config['main_coil'].pulse()
            self.config['hold_coil'].enable()
        else:
            self.config['main_coil'].enable()

    @event_handler(5)
    def event_sw_release(self, **kwargs):
        """Handle sw_release control event."""
        del kwargs
        self.sw_release()

    def sw_release(self):
        """Deactivate the flipper via software as if the flipper button was released.

        See the documentation for sw_flip() for details.
        """
        self._sw_flipped = False

        # disable the flipper coil(s)
        self.config['main_coil'].disable()

        if self.config['hold_coil']:
            self.config['hold_coil'].disable()

    def _ball_search(self, phase, iteration):
        del phase
        del iteration
        self.sw_flip()
        self.machine.delay.add(self.config['ball_search_hold_time'],
                               self.sw_release,
                               'flipper_{}_ball_search'.format(self.name))
        return True



"""Post an event to trigger an eject."""
from typing import List

from mpf.devices.ball_device.ball_device_ejector import BallDeviceEjector


class EventEjector(BallDeviceEjector):

    """Post an event to trigger an eject."""

    __slots__ = []  # type: List[str]

    def __init__(self, config, ball_device, machine):
        """Initialise ejector."""
        super().__init__(config, ball_device, machine)
        self.config = self.machine.config_validator.validate_config("ball_device_ejector_event", self.config)

    async def eject_one_ball(self, is_jammed, eject_try, balls_in_device):
        """Post event."""
        for event in self.config["events_when_eject_try"]:
            self.machine.events.post(event, is_jammed=is_jammed, eject_try=eject_try, balls_in_device=balls_in_device)

    async def reorder_balls(self):
        """Reorder balls when jammed."""
        for event in self.config["events_when_reoder_balls"]:
            self.machine.events.post(event)

    def ball_search(self, phase, iteration):
        """Run ball search."""
        for event in self.config["events_when_ball_search"]:
            self.machine.events.post(event, phase=phase, iteration=iteration)

        # do not wait for this device
        return False



"""Contains the BallController class which manages and tracks all the balls in a pinball machine."""

import asyncio
from typing import Union, Iterable, Optional

from mpf.devices.ball_device.ball_device import BallDevice

from mpf.core.delays import DelayManager
from mpf.core.machine import MachineController
from mpf.core.utility_functions import Util
from mpf.core.mpf_controller import MpfController


class BallController(MpfController):

    """Tracks and manages all the balls in a pinball machine."""

    config_name = "ball_controller"

    __slots__ = ["delay", "num_balls_known", "_add_new_balls_task", "_captured_balls"]

    def __init__(self, machine: MachineController) -> None:
        """Initialise ball controller.

        Args:
        ----
            machine : :class:`MachineController`. A reference to the instance
                of the MachineController object.

        """
        super().__init__(machine)

        self.delay = DelayManager(self.machine)

        self.num_balls_known = 0

        # register for events
        self.machine.events.add_handler('request_to_start_game',
                                        self.request_to_start_game)
        self.machine.events.add_handler('init_phase_2',
                                        self._initialize)
        self.machine.events.add_handler('init_phase_4',
                                        self._init4, priority=100)

        self.machine.events.add_handler('shutdown',
                                        self._stop)

        self._add_new_balls_task = None                                         # type: Optional[asyncio.Task]
        self._captured_balls = asyncio.Queue()      # type: asyncio.Queue

    def _init4(self, **kwargs):
        del kwargs

        # see if there are non-playfield devices
        for device in self.machine.ball_devices.values():
            if device.is_playfield():
                continue
            # found a device
            break
        else:
            # there is no non-playfield device. end this.
            return

        self._add_new_balls_task = self.machine.clock.loop.create_task(self._add_new_balls_to_playfield())
        self._add_new_balls_task.add_done_callback(Util.raise_exceptions)

    def _stop(self, **kwargs) -> None:
        del kwargs
        if self._add_new_balls_task:
            self._add_new_balls_task.cancel()
            self._add_new_balls_task = None

    def _get_total_balls_in_devices(self) -> int:
        """Return the total ball count over all devices in the machine."""
        balls = 0
        # get count for all ball devices
        for device in self.machine.ball_devices.values():
            if device.is_playfield():
                continue

            balls += device.ball_count_handler.counter.count_balls_sync()
        return balls

    def add_captured_ball(self, source: BallDevice) -> None:
        """Inform ball controller about a captured ball (which might be new)."""
        self._captured_balls.put_nowait(source)

    async def _add_new_balls_to_playfield(self) -> None:
        # initial count
        self.num_balls_known = await self._count_all_balls_in_devices()
        self.info_log("Initial balls found: %s", self.num_balls_known)

        while True:
            capture = await self._captured_balls.get()
            balls = await self._count_all_balls_in_devices()
            if balls > self.num_balls_known:
                self.num_balls_known += 1
                capture.balls += 1
                capture.available_balls += 1
                self.info_log("Found a new ball which was captured from %s. Known balls: %s", capture.name,
                              self.num_balls_known)
                self.machine.events.post("found_new_ball")

            if len(self.machine.playfields) > 1:
                self._balance_playfields()

    def _balance_playfields(self) -> None:
        # find negative counts
        for playfield_target in self.machine.playfields.values():
            if playfield_target.balls < 0:
                for playfield_source in self.machine.playfields.values():
                    if playfield_source.balls > 0:
                        playfield_source.balls -= 1
                        playfield_source.available_balls -= 1
                        playfield_target.balls += 1
                        playfield_target.available_balls += 1
                        # post event
                        self.machine.events.post("playfield_jump", source=playfield_source, target=playfield_target)
                        break

    async def _count_all_balls_in_devices(self) -> int:
        """Count balls in all devices."""
        while True:
            # wait until all devices are stable
            # prepare futures in case we have to wait to prevent race between count and building futures
            futures = []
            for device in self.machine.ball_devices.values():
                if not device.is_playfield():
                    futures.append(asyncio.ensure_future(device.ball_count_handler.counter.wait_for_count_stable()))

            try:
                return self._get_total_balls_in_devices()
            except ValueError:
                await asyncio.wait(futures)
                continue

    def _count_balls(self) -> int:
        """Count balls in devices and raise ValueError if counts are not stable."""
        self.debug_log("Counting Balls")
        balls = 0

        for device in self.machine.ball_devices.values():
            # skip playfields
            if device.is_playfield():
                continue
            if device.ball_count_handler.counter.config.get('ball_switches'):
                for switch in device.ball_count_handler.counter.config['ball_switches']:
                    if self.machine.switch_controller.is_active(
                            switch, ms=device.ball_count_handler.counter.config['entrance_count_delay']):
                        balls += 1
                    elif self.machine.switch_controller.is_inactive(
                            switch, ms=device.ball_count_handler.counter.config['exit_count_delay']):
                        continue
                    else:
                        raise ValueError("switches not stable")
            elif 'trough' in device.tags:
                # special handling for troughs (needed for gottlieb)
                balls += device.balls

        return balls

    def _initialize(self, **kwargs) -> None:
        """Initialise ball controller."""
        # If there are no ball devices, then the ball controller has no work to
        # do and will create errors, so we just abort.
        del kwargs
        if not hasattr(self.machine, 'ball_devices'):
            return

        for device in self.machine.ball_devices.values():
            priority = 0
            if 'drain' in device.tags or 'trough' in device.tags:  # device is used to drain balls from pf
                priority += 1   # order handlers
                self.machine.events.add_handler('balldevice_' + device.name +
                                                '_ball_enter',
                                                self._ball_drained_handler, priority=20 + priority)

    def dump_ball_counts(self) -> None:
        """Dump ball count of all devices."""
        self.info_log("Known balls: %s", self.num_balls_known)
        for device in self.machine.ball_devices.values():
            self.info_log("%s contains %s balls. Tags %s", device.name, device.balls, device.tags)

    def request_to_start_game(self, **kwargs) -> bool:
        """Handle result of the *request_to_start_game* event.

        Checks to make sure that the balls are in all the right places and
        returns. If too many balls are missing (based on the config files 'Min
        Balls' setting), it will return False to reject the game start request.
        """
        del kwargs
        try:
            balls = self._count_balls()
        except ValueError:
            balls = -1
        self.debug_log("Received request to start game.")
        if balls < self.machine.config['machine']['min_balls']:
            self.dump_ball_counts()
            self.warning_log("BallController denies game start. Not enough "
                             "balls. %s found. %s required", balls, self.machine.config['machine']['min_balls'])

            self.info_log("Triggering ball search.")
            for playfield in self.machine.playfields.values():
                playfield.ball_search.enable()
                playfield.ball_search.start()
            return False

        if self.machine.config['game']['allow_start_with_ball_in_drain']:
            allowed_positions = ['home', 'trough', 'drain']
        else:
            allowed_positions = ['home', 'trough']

        if (not self.machine.config['game']['allow_start_with_loose_balls'] and
                not self.are_balls_collected(allowed_positions)):
            self.collect_balls('home')
            self.dump_ball_counts()
            self.warning_log("BallController denies game start. Balls are not "
                             "in their home positions.")
            return False

        # allow start
        return True

    def are_balls_collected(self, target: Union[str, Iterable[str]]) -> bool:
        """Check to see if all the balls are contained in devices tagged with the parameter that was passed.

        Note if you pass a target that's not used in any ball devices, this
        method will return True. (Because you're asking if all balls are
        nowhere, and they always are. :)

        Args:
        ----
            target: String or list of strings of the tags you'd like to
                collect the balls to. Default of None will be replaced with
                'home' and 'trough'.
        """
        self.debug_log("Checking to see if all the balls are in devices tagged"
                       " with '%s'", target)

        if isinstance(target, str):
            target = Util.string_to_event_list(target)

        count = 0
        devices = set()

        for tag in target:
            for device in self.machine.ball_devices.items_tagged(tag):      # type: ignore
                devices.add(device)

        if not devices:
            # didn't find any devices matching that tag, so we return True
            return True

        for device in devices:
            count += device.balls
            self.debug_log('Found %s ball(s) in %s. Found %s total',
                           device.balls, device.name, count)

        if count == self.machine.ball_controller.num_balls_known:
            self.debug_log("Yes, all balls are collected")
            return True

        self.debug_log("No, all balls are not collected. Balls Counted: %s. "
                       "Total balls known: %s", count,
                       self.machine.ball_controller.num_balls_known)
        return False

    def collect_balls(self, target='home, trough') -> None:
        """Ensure that all balls are in contained in ball devices with the tag or list of tags you pass.

        Typically this would be used after a game ends, or when the machine is
        reset or first starts up, to ensure that all balls are in devices
        tagged with 'home' and/or 'trough'.

        Args:
        ----
            target: A string of the tag name or a list of tags names of the
                ball devices you want all the balls to end up in. Default is
                ['home', 'trough'].

        """
        tag_list = Util.string_to_event_list(target)

        self.debug_log("Collecting all balls to devices with tags '%s'",
                       tag_list)

        target_devices = set()
        source_devices = set()
        balls_to_collect = False

        for tag in tag_list:
            for device in self.machine.ball_devices.items_tagged(tag):      # type: ignore
                target_devices.add(device)

        for device in self.machine.ball_devices.values():
            if device not in target_devices:
                if device.available_balls:
                    source_devices.add(device)
                    balls_to_collect = True

        if balls_to_collect:
            self.debug_log("Ejecting all balls from: %s", source_devices)

            self.machine.events.post('collecting_balls')
            '''event: collecting_balls

            desc: Posted by the ball controller when it starts the collecting
                balls process.

            '''

            for device in target_devices:
                self.machine.events.remove_handler(self._collecting_balls_entered_callback)
                self.machine.events.add_handler(
                    'balldevice_{}_ball_enter'.format(device.name),
                    self._collecting_balls_entered_callback,
                    target=target, priority=10)

            for device in source_devices:
                if not device.is_playfield():
                    if "drain" in device.tags:
                        device.eject_all(device.find_next_trough())
                    else:
                        device.eject_all()
        else:
            self.debug_log("All balls are collected")
            self._collecting_balls_complete()

    def _collecting_balls_entered_callback(self, target: str, new_balls: int, unclaimed_balls: int, **kwargs) -> dict:
        del kwargs
        del new_balls
        if self.are_balls_collected(target=target):
            self._collecting_balls_complete()

        return {'unclaimed_balls': unclaimed_balls}

    def _collecting_balls_complete(self) -> None:
        self.machine.events.remove_handler(self._collecting_balls_entered_callback)
        self.machine.events.post('collecting_balls_complete')
        '''event: collecting_balls_complete

        desc: Posted by the ball controller when it has finished the collecting
            balls process.

        '''

    def _ball_drained_handler(self, new_balls: int, unclaimed_balls: int, device: BallDevice, **kwargs) -> None:
        del kwargs
        del new_balls
        self.machine.events.post_relay('ball_drain',
                                       device=device,
                                       balls=unclaimed_balls)
        '''event: ball_drain

        desc: A ball (or balls) has just drained. (More specifically, ball(s)
        have entered a ball device tagged with "drain".)

        This is a relay event.

        args:

        device: The ball device object that received the ball(s)

        balls: The number of balls that have just drained. Any balls remaining
        after the relay will be processed as newly-drained balls.

        '''



"""Contains the DelayManager base classes."""

import uuid
from functools import partial
from typing import Any, Callable, Dict, Tuple, Union
from mpf.core.mpf_controller import MpfController

MYPY = False
if MYPY:    # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import

__api__ = ['DelayManager']


class DelayManager(MpfController):

    """Handles delays for one object.

    By default, a machine-wide instance is created and available via
    ``self.machine.delay``.

    Individual modes also have Delay Managers which can be accessed in
    mode code via ``self.delay``. (Delays in mode-based delay managers
    are automatically removed when the mode stops.)

    """

    __slots__ = ["delays", "registry"]

    config_name = "delay_manager"

    def __init__(self, machine: "MachineController") -> None:
        """Initialise delay manager."""
        self.delays = {}        # type: Dict[str, Tuple[Any, Callable]]
        super().__init__(machine)

    def add(self, ms: int, callback: Callable[..., None], name: str = None,
            **kwargs) -> str:
        """Add a delay.

        Args:
        ----
            ms: The number of milliseconds you want this delay to be for.
            callback: The method that is called when this delay ends.
            name: String name of this delay. This name is arbitrary and only
                used to identify the delay later if you want to remove or
                change it. If you don't provide it, a UUID4 name will be
                created.
            **kwargs: Any other (optional) kwarg pairs you pass will be
                passed along as kwargs to the callback method.

        Returns string name or UUID4 of the delay which you can use to remove it
        later.
        """
        if not name:
            name = str(uuid.uuid4())
        self.debug_log("Adding delay. Name: '%s' ms: %s, callback: %s, "
                       "kwargs: %s", name, ms, callback, kwargs)

        try:
            delay = self.delays.pop(name)
        except KeyError:
            pass
        else:
            self.machine.clock.unschedule(delay[0])

        self.delays[name] = (self.machine.clock.schedule_once(
            partial(self._process_delay_callback, name, callback, **kwargs),
            ms / 1000.0), callback)

        return name

    def remove(self, name: str):
        """Remove a delay by name.

        Removing a delay prevents the callback from being called and cancels
        the delay.

        Args:
        ----
            name: String name of the delay you want to remove. If there is no
                delay with this name, that's ok. Nothing happens.
        """
        self.debug_log("Removing delay: '%s'", name)
        try:
            delay = self.delays.pop(name)
        except KeyError:
            pass
        else:
            self.machine.clock.unschedule(delay[0])

    def add_if_doesnt_exist(self, ms: int, callback: Callable[..., None],
                            name: str, **kwargs) -> str:
        """Add a delay only if a delay with that name doesn't exist already.

        Args:
        ----
            ms: Int of the number of milliseconds you want this delay to be
                for.
            callback: The method that is called when this delay ends.
            name: String name of this delay. This name is arbitrary and only
                used to identify the delay later if you want to remove or
                change it.
            **kwargs: Any other (optional) kwarg pairs you pass will be
                passed along as kwargs to the callback method.

        Returns string name of the delay which you can use to remove it later.
        """
        if not self.check(name):
            return self.add(ms, callback, name, **kwargs)

        return name

    def check(self, delay: str) -> bool:
        """Check to see if a delay exists.

        Args:
        ----
            delay: A string of the delay you're checking for.

        Returns true if the delay exists. False otherwise.
        """
        return delay in self.delays

    def reset(self, ms: int, callback: Callable[..., None], name: str,
              **kwargs) -> str:
        """Reset a delay.

        Resetting will first delete the existing delay (if it exists) and then
        add new delay with the new settings. If the delay does not exist,
        that's ok, and this method is essentially the same as just adding a
        delay with this name.

        Args:
        ----
            ms: The number of milliseconds you want this delay to be for.
            callback: The method that is called when this delay ends.
            name: String name of this delay. This name is arbitrary and only
                used to identify the delay later if you want to remove or
                change it. If you don't provide it, a UUID4 name will be
                created.
            **kwargs: Any other (optional) kwarg pairs you pass will be
                passed along as kwargs to the callback method.

        Returns string name or UUID4 of the delay which you can use to remove it
        later.
        """
        if name in self.delays:
            self.remove(name)

        return self.add(ms, callback, name, **kwargs)

    def clear(self) -> None:
        """Remove (clear) all the delays associated with this DelayManager."""
        for name in list(self.delays.keys()):
            self.machine.clock.unschedule(self.delays[name][0])
            self.remove(name)

        self.delays = {}

    def run_now(self, name: str):
        """Run a delay callback now instead of waiting until its time comes.

        This will cancel the future running of the delay callback.

        Args:
        ----
            name: Name of the delay to run. If this name is not an active
                delay, that's fine. Nothing happens.
        """
        if name in self.delays:
            try:
                # have to save the callback ref first, since if the callback
                # schedules a new delay with the same name, then the removal
                # will remove it
                cb = self.delays[name][1]
                self.remove(name)
                cb()
            except KeyError:
                pass

    def _process_delay_callback(self, name: str, callback: Callable[..., None], **kwargs):
        # Process the delay callback and run the event queue afterwards
        self.debug_log("---Processing delay: %s", name)
        try:
            del self.delays[name]
        except KeyError:
            pass
        callback(**kwargs)
        self.machine.events.process_event_queue()



import logging
import asyncio
import grpc
from grpc.experimental import aio

from mpf.platforms.visual_pinball_engine import platform_pb2_grpc
from mpf.platforms.visual_pinball_engine import platform_pb2


class MpfHardwareService(platform_pb2_grpc.MpfHardwareServiceServicer):

    def Start(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendSwitchChanges(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Quit(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

async def switch_generator():
    state = True
    while True:
        print("+")
        state = not state
        change = platform_pb2.SwitchChanges()
        change.switch_number = "0"
        change.switch_state = state
        yield change
        await asyncio.sleep(1)

async def connect():
    channel = aio.insecure_channel("localhost:50051")
    await channel.channel_ready()
    stub = platform_pb2_grpc.MpfHardwareServiceStub(channel)
    configuration = platform_pb2.MachineConfiguration(
            known_switches_with_initial_state={"0": True, "3": False, "6": False},
            known_lights=["light-0", "light-1"],
            known_coils=["0", "1", "2"])
    command_stream = stub.Start(configuration)
    asyncio.ensure_future(stub.SendSwitchChanges(switch_generator()))
    while True:
        command = await command_stream.read()
        print(command)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    asyncio.run(connect())



"""Fast DMD support."""
from mpf.platforms.interfaces.dmd_platform import DmdPlatformInterface


class FASTDMD(DmdPlatformInterface):

    """Object for a FAST DMD."""

    def __init__(self, machine, sender):
        """Initialise DMD."""
        self.machine = machine
        self.send = sender

        # Clear the DMD
        # todo

    def set_brightness(self, brightness: float):
        """Set brightness."""
        # not supported
        assert brightness == 1.0

    def update(self, data: bytes):
        """Update data on the DMD.

        Args:
        ----
            data: bytes to send to DMD
        """
        self.send(data)



"""Contains the Match mode code."""
import random

import math

from mpf.core.settings_controller import SettingEntry

from mpf.core.async_mode import AsyncMode


class Match(AsyncMode):

    """Match mode."""

    __slots__ = []

    def __init__(self, *args, **kwargs):
        """Initialise match mode."""
        super().__init__(*args, **kwargs)
        # add setting
        self.machine.settings.add_setting(SettingEntry("match_percentage", "Match percentage", 500,
                                                       "match_percentage", 10,
                                                       {0: "off", 2: "2%", 5: "5%", 10: "10%", 15: "15%", 30: "30%",
                                                        50: "50%"}))

    def _get_match_numbers(self):
        """Calculate match numbers."""
        match_numbers = []
        for player in self.machine.game.player_list:
            match_numbers.append(player.score % 100)
        return match_numbers

    def _get_winner_number(self, match_numbers, match_percentage) -> int:
        """Find the winning number.

        Return the winning number.
        """
        # check if a player will win
        if random.randint(0, 100) < math.pow(match_percentage, len(set(match_numbers))):
            # we got a winner
            return random.choice(match_numbers)

        step = self.config.get("mode_settings", {}).get("non_match_number_step", 1)

        # no winner. return some other number
        non_winning_numbers = list(set(range(0, 100, step)) - set(match_numbers))
        return random.choice(non_winning_numbers)

    async def _run(self) -> None:
        """Run match mode."""
        # no player, no match
        if not self.machine.game or not self.machine.game.player_list:
            return

        match_percentage = self.machine.settings.get_setting_value("match_percentage")

        match_numbers = self._get_match_numbers()
        winner_number = self._get_winner_number(match_numbers, match_percentage)
        winners = match_numbers.count(winner_number)

        event_args = {
            "winner_number": winner_number,
            "winners": winners
        }

        for i in range(0, self.machine.game.max_players):
            event_args["match_number{}".format(i)] = match_numbers[i] if len(match_numbers) > i else ""

        if not winners:
            # no winner
            await self.machine.events.post_queue_async("match_no_match", **event_args)
            '''event: match_no_match

            desc: All players missed the match number.

            args:
              winner_number: Winner number
              winners: Number of winners (always 0 here)
              match_number0: Match number for player 0
              match_number1: Match number for player 1
              match_numberX: Match number for player X (up to max players)
            '''
        else:
            # we got a winner
            await self.machine.events.post_queue_async("match_has_match", **event_args)
            '''event: match_has_match

            desc: At least one player has a match.

            args:
              winner_number: Winner number
              winners: Number of winners (always more than 0 here)
              match_number0: Match number for player 0
              match_number1: Match number for player 1
              match_numberX: Match number for player X (up to max players)
            '''
        # that is it. credits mode should hook into the match_has_match event and award credits



import time

from mpf.core.logging import LogMixin

from mpf.tests.MpfGameTestCase import MpfGameTestCase


class BenchmarkSwitchHits(MpfGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'benchmarks/machine_files/switch_hits/'

    def get_platform(self):
        return 'virtual'

    def setUp(self):
        LogMixin.unit_test = False
        super().setUp()

    def _output(self, start, end, end2, num):
        print("Duration {:.5f}ms Processing {:.5f}ms Total: {:5f}ms Per second: {:2f}".format(
            (1000 * (end - start) / num), ((end2 - end) * 1000) / num, (1000 * (end2 - start)) / num,
            (num * 1) / (end2 - start)
        ))

    def testMinimalSwitch(self):
        for i in range(1000):
            self.machine.switch_controller.process_switch_by_num("4", 1, self.machine.default_platform)
            self.machine.switch_controller.process_switch_by_num("4", 0, self.machine.default_platform)
        self.advance_time_and_run()

        num = 10000
        for runs in range(10):
            start = time.time()
            for i in range(num):
                self.machine.switch_controller.process_switch_by_num("4", 1, self.machine.default_platform)
                self.machine.switch_controller.process_switch_by_num("4", 0, self.machine.default_platform)
            end = time.time()
            self.advance_time_and_run()
            end2 = time.time()
            self._output(start, end, end2, num)

    def testSimpleHandlers(self):
        hit = 0
        handler = 100
        for i in range(handler):
            self.machine.switch_controller.add_switch_handler("s_switch4", lambda: hit.__add__(1))

        for i in range(1000):
            self.machine.switch_controller.process_switch_by_num("4", 1, self.machine.default_platform)
            self.machine.switch_controller.process_switch_by_num("4", 0, self.machine.default_platform)
        self.advance_time_and_run()

        num = 10000
        for runs in range(10):
            start = time.time()
            for i in range(num):
                self.machine.switch_controller.process_switch_by_num("4", 1, self.machine.default_platform)
                self.machine.switch_controller.process_switch_by_num("4", 0, self.machine.default_platform)
            end = time.time()
            self.advance_time_and_run()
            end2 = time.time()
            self._output(start, end, end2, num)

    def testTimedSwitchHandlers(self):
        hit = 0
        handler = 1
        for i in range(handler):
            self.machine.switch_controller.add_switch_handler("s_switch4", lambda: hit.__add__(1), ms=100)
        for i in range(1000):
            self.machine.switch_controller.process_switch_by_num("4", 1, self.machine.default_platform)
            self.machine.switch_controller.process_switch_by_num("4", 0, self.machine.default_platform)
        self.advance_time_and_run()

        num = 10000
        for runs in range(10):
            start = time.time()
            for i in range(num):
                self.machine.switch_controller.process_switch_by_num("4", 1, self.machine.default_platform)
                self.machine.switch_controller.process_switch_by_num("4", 0, self.machine.default_platform)
            end = time.time()
            self.advance_time_and_run()
            end2 = time.time()
            self._output(start, end, end2, num * handler)

    def testBenchmarkIgnoreWindowMsHits(self):
        for i in range(1000):
            self.machine.switch_controller.process_switch_by_num("3", 1, self.machine.default_platform)
            self.machine.switch_controller.process_switch_by_num("3", 0, self.machine.default_platform)
        self.machine.switch_controller.process_switch_by_num("3", 1, self.machine.default_platform)
        self.advance_time_and_run()

        num = 10000
        for runs in range(10):
            start = time.time()
            for i in range(num):
                self.machine.switch_controller.process_switch_by_num("3", 1, self.machine.default_platform)
                self.machine.switch_controller.process_switch_by_num("3", 0, self.machine.default_platform)
            end = time.time()
            self.machine.switch_controller.process_switch_by_num("3", 1, self.machine.default_platform)
            self.advance_time_and_run()
            end2 = time.time()
            self._output(start, end, end2, num)

    def testPlayfieldActiveSwitchHits(self):
        for i in range(1000):
            self.machine.switch_controller.process_switch_by_num("1", 1, self.machine.default_platform)
            self.machine.switch_controller.process_switch_by_num("1", 0, self.machine.default_platform)
        self.machine_run()

        num = 10000
        for runs in range(10):
            start = time.time()
            for i in range(num):
                self.machine.switch_controller.process_switch_by_num("1", 1, self.machine.default_platform)
                self.machine.switch_controller.process_switch_by_num("1", 0, self.machine.default_platform)
            end = time.time()
            self.machine_run()
            end2 = time.time()
            self._output(start, end, end2, num)

    def test20TagsHits(self):
        for i in range(1000):
            self.machine.switch_controller.process_switch_by_num("2", 1, self.machine.default_platform)
            self.machine.switch_controller.process_switch_by_num("2", 0, self.machine.default_platform)
        self.machine_run()

        for runs in range(10):
            num = 10000
            start = time.time()
            for i in range(num):
                self.machine.switch_controller.process_switch_by_num("2", 1, self.machine.default_platform)
                self.machine.switch_controller.process_switch_by_num("2", 0, self.machine.default_platform)
            end = time.time()
            self.machine_run()
            end2 = time.time()
            self._output(start, end, end2, num)



from mpf._version import log_url

from mpf.tests.MpfTestCase import MpfTestCase, test_config, test_config_directory


class TestConfigErrors(MpfTestCase):

    def setUp(self):
        self.save_and_prepare_sys_path()

    def tearDown(self):
        self.restore_sys_path()

    @test_config_directory("tests/machine_files/config_errors/broken_show")
    @test_config("show.yaml")
    def test_show_player_in_show(self):
        with self.assertRaises(AssertionError) as e:
            super().setUp()
            self.post_event("play_broken_show")
        self.assertEqual(str(e.exception),
                         'Config File Error in show: "broken_show" >> Invalid section "light_player:" '
                         'found. Did you mean "lights:" instead? Context: broken_show '
                         'Error Code: CFE-show-3 ({})'.format(log_url.format("CFE-show-3")))
        self.loop.close()
